"""Conversion utilities for ParaDime.

The :mod:`paradime.utils.convert` module implements various conversion
functions for tensors-like objects and index lists.
"""

import functools
from typing import Union

import numpy as np
import torch

from paradime.types import TensorLike


def to_numpy(X: Union[TensorLike, list[float]]) -> np.ndarray:
    """Converts a tensor-like object to a numpy array.

    Args:
        X: The tensor-like object to be converted.

    Returns:
        The resulting numpy array.
    """
    if isinstance(X, torch.Tensor):
        return X.detach().cpu().numpy()
    elif isinstance(X, np.ndarray):
        return X
    elif isinstance(X, list):
        return np.array(X)
    else:
        raise TypeError(f"Input type {type(X)} not supported")


def to_torch(X: Union[TensorLike, list[float]]) -> torch.Tensor:
    """Converts a tensor-like object to a PyTorch tensor.

    Args:
        X: The tensor-like object to be converted.

    Returns:
        The resulting PyTorch tensor. If the input was not a PyTorch tensor
        already, the output tensor will be of type float32 for float inputs and
        of type int32 for integer inputs.
    """
    if isinstance(X, torch.Tensor):
        return X
    elif isinstance(X, (np.ndarray, list)):
        tensor = torch.tensor(X)
        if tensor.is_floating_point():
            return tensor.float()
        elif not tensor.is_complex():
            return tensor.int()
        else:
            return tensor
    else:
        raise TypeError(f"Input type {type(X)} not supported")


@functools.cache
def rowcol_to_triu_index(i: int, j: int, dim: int) -> int:
    """Converts matrix indices to upper-triangular form.

    Converts a pair of row and column indices of a symmetrical square array to
    the corresponding index of the list of upper triangular values.

    Args:
        i: The row index.
        j: The column index.
        dim: The size of the square matrix.

    Returns:
        The upper triangular index.

    Raises:
        ValueError: For diagonal indices (i.e., if i equals j).
    """
    if i < j:
        index = np.round(i * (dim - 1.5) + j - i**2 * 0.5 - 1)
        return int(index)
    elif i > j:
        return rowcol_to_triu_index(j, i, dim)
    else:
        raise ValueError(
            "Indices of diagonal elements cannot be converted to "
            "upper-triangular form."
        )


@functools.cache
def triu_to_square_dim(len_triu: int) -> int:
    """Calculates the size of a square matrix given the length of the list of
    its upper-triangular values.

    Args:
        len_triu: The lenght of the list of upper-triangular values.

    Returns:
        The size of the square matrix.
    """
    return int(np.ceil(np.sqrt(len_triu * 2)))



"""Plotting utilities for ParaDime.

The :mod:`paradime.utils.plotting` module implements plotting functions and
color palette retrieval.
"""

from typing import Any, Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:
    import matplotlib.colors
    import matplotlib.axes

import numpy as np

from paradime.types import TensorLike
from paradime.utils.make_palette import make_palette

def get_color_palette() -> dict[str, str]:
    """Get the custom ParaDime color palette.
    
    The palette is usually located in an assets folder in the form of a JSON
    file. If the JSON file is not found, this method attemps to create it from
    parsing an SVG file.

    Returns:
        The color palette as a dict of names and hex color values.

    Raises:
        FileNotFoundError: If neither the JSON nor the SVG file can be found.
    """
    import json
    import os

    utils_path = os.path.dirname(__file__)
    json_path = os.path.join(utils_path, 'palette.json')
    svg_path = os.path.join(utils_path, 'palette.svg')

    if not os.path.isfile(json_path):
        if os.path.isfile(svg_path):
            make_palette(svg_path, json_path)
        else:
            raise FileNotFoundError(
                "Could not find JSON or SVG file to create/import palette."
            )
    with open(json_path, 'r') as f:
            return json.load(f)

def get_colormap() -> "matplotlib.colors.ListedColormap":
    from ._cmap import _paradime_cmap
    return _paradime_cmap

def scatterplot(
    coords: TensorLike,
    labels: Optional[TensorLike] = None,
    colormap: Optional[list[str]] = None,
    labels_to_index: Optional[dict] = None,
    figsize: tuple[float, float] = (10,10),
    bgcolor: Optional[str] = "#fcfcfc",
    legend: bool = True,
    legend_options: Optional[dict[str, Any]] = None,
    ax: Optional["matplotlib.axes.Axes"] = None,
    **kwargs
) -> None:
    """Creates a scatter plot of points at the given coordinates.

    Args:
        coords: The coordinates of the points.
        labels: An list of categorical labels. If labels are given,
            a categorical color scale is used and a legend is constructed
            automatically.
        colormap: A list of colors to use instead of the default categorical
            color scale based on the ParaDime palette.
        labels_to_index: A dict that maps labels to indices which are then used
            to access the colors in the categorical color scale.
        figsize: Width and height of the plot in inches.
        bgcolor: The background color of the plot, which by default is also
            to draw thin outlines around the points.
        legend: Whether or not to include the automatically created legend.
        legend_options: A dict of keyword arguments that are passed on to the
            legend method.
        ax: An axes of the current figure. This argument is useful if the
            scatterplot should be added to an existing figure.
        kwargs: Any other keyword arguments are passed on to matplotlib's
            `scatter` method.

    Returns:
        The :class:`matplotlib.axes.Axes` instance of the plot.
    """

    from matplotlib import pyplot as plt
    from matplotlib import patches
    
    if colormap is None:
        colormap = list(get_color_palette().values())

    colors: Union[str, list[str]]
    if labels is None:
        colors = colormap[0]
    else:
        unique, indices = np.unique(labels, return_inverse=True)
        if labels_to_index is not None:
            indices = [ labels_to_index[i] for i in labels ]
        colors = [ colormap[i] for i in indices ]

        def rect(color: str) -> patches.Rectangle:
            return patches.Rectangle((0.,0.), 1., 1., fc=color)

        rects = [ rect(colormap[i]) for i in np.unique(indices) ]

    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(1,1,1)
    else:
        assert isinstance(ax, plt.Axes)
        fig = ax.get_figure()

    dpi: float = fig.dpi

    if hasattr(kwargs, 's'):
        pointsize = kwargs['s']
    else:
        min_figsize = min(figsize)
        pointwidth = min(
            max(
                0.005 * min_figsize * dpi,
                min_figsize * dpi / np.sqrt(len(coords)) * 0.2
            ),
            0.02 * min_figsize * dpi
        )
        pointsize = pointwidth**2

    ax.set_facecolor(bgcolor)

    scatter_kwargs = {
        'c': colors,
        's': pointsize,
        'edgecolor': bgcolor,
    }
    if kwargs:
        scatter_kwargs = {**scatter_kwargs, **kwargs}

    points = ax.scatter(
        x=coords[:,0],
        y=coords[:,1],
        **scatter_kwargs
    )
    ax.set_aspect('equal')
    ax.set_axis_off()
    points.set_linewidths(.08 * np.sqrt(pointsize))

    if legend and labels is not None:
        legend_kwargs = {
            'handleheight': 1.,
            'handlelength': 1.,
            'loc': 4,
            'fancybox': False,
        }
        if legend_options is not None:
            legend_kwargs = {**legend_kwargs, **legend_options}
        
        ax.legend(rects, unique, **legend_kwargs)

    return ax
    


"""Losses for ParaDime routines.

The :mod:`paradime.loss` module implements the specification of losses for
ParaDime routines. The supported losses are
:class:`paradime.loss.RelationLoss`,
:class:`paradime.loss.ClassificationLoss`,
:class:`paradime.loss.ReconstructionLoss`, and
:class:`paradime.loss.CompoundLoss`.
"""

from typing import Optional, Literal, Union

import numpy as np
import torch

from paradime import exceptions
from paradime import models
from paradime import relationdata
from paradime import relations
from paradime.types import BinaryTensorFun
from paradime import utils


class Loss(torch.nn.Module):
    """Base class for losses.

    Custom losses should subclass this class.

    Attributes:
        name: The name of the loss (used by logging functions).
    """

    _prefix = "loss"

    def __init__(self, name: Optional[str] = None):
        super().__init__()

        if name is None:
            self._name = self._prefix + "_" + str(id(self))
        else:
            self._name = name

        self._accumulated: float = 0.0
        self.history: list[float] = []

        self._history_hook = self.register_forward_hook(self._add_last)

    @property
    def name(self) -> str:
        return self._name

    def __call__(self, *args, **kwargs) -> torch.Tensor:
        # redefine to call super to improve type hinting
        return super().__call__(*args, **kwargs)

    def _add_last(
        self,
        model: torch.nn.Module,
        input: torch.Tensor,
        output: torch.Tensor,
    ) -> None:
        self._accumulated += float(output.detach().cpu().item())

    def checkpoint(self) -> None:
        """Create a checkpoint of the most recent accumulated loss.

        Appends the value of the most recent accumulated loss to the loss's
        ``history`` attribute. If the loss is a
        :class:`paradime.loss.CompoundLoss`, checkpoints are also created for
        each individual loss.
        """
        self.history.append(self._accumulated)
        self._accumulated = 0.0

    def forward(
        self,
        model: models.Model,
        global_relations: dict[str, relationdata.RelationData],
        batch_relations: dict[str, relations.Relations],
        batch: dict[str, torch.Tensor],
        device: torch.device,
    ) -> torch.Tensor:
        """Apply the loss to a batch of input data.

        Args:
            model: The :class:`torch.nn.module` used to embed, classify, or
                reconstruct a batch of input data.
            global_relations: A dictionary with
                :class:`paradime.relationdata.RelationData` computed for the
                whole dataset.
            batch_relations: A dictionary with
                :class:`paradime.relations.Relations` to be computed for the
                batch of input data.
            batch: A batch of input data as a dictionary of PyTorch tensors.
            device: The device that all relevant tensors will be moved to.

        Returns:
            A single-item PyTorch tensor with the computed loss.
        """
        raise NotImplementedError()


class RelationLoss(Loss):
    """A loss that compares batch-wise relation data against a subset of
    global relation data.

    This loss applies a specified loss function to a subset of pre-computed
    global relations and the batch-wise relations found under specified keys,
    respectively. Batch-wise relations are computed from embedded coordinates
    by applying the model's :meth:`embed` method to the specified data entry
    of the input batch.

    Args:
        loss_function: The loss function to be applied.
        global_relation_key: Key under which to find the global relations.
        batch_relation_key: Key under which to find the batch-wise relations.
        embedding_method: The model method to be used for embedding the batch
            of input data.
        name: Name of the loss (used by logging functions).
    """

    _prefix = "rel_loss"

    def __init__(
        self,
        loss_function: BinaryTensorFun,
        global_relation_key: str = "rel",
        batch_relation_key: str = "rel",
        embedding_method: str = "embed",
        normalize_sub: bool = True,
        name: Optional[str] = None,
    ):
        super().__init__(name)

        self.loss_function = loss_function
        self.global_relation_key = global_relation_key
        self.batch_relation_key = batch_relation_key
        self.embedding_method = embedding_method
        self.normalize_sub = normalize_sub

        self._use_from_to = False

    def _check_sampling_and_relations(
        self,
        sampling: Literal["standard", "negative_edge"],
        batch_relations: dict[str, relations.Relations],
    ) -> None:
        if isinstance(
            batch_relations[self.batch_relation_key], relations.DistsFromTo
        ):
            if sampling == "negative_edge":
                self._use_from_to = True
                self.data_key = "from_to_data"
            else:
                raise exceptions.UnsupportedConfigurationError(
                    "RelationLoss does not support DistsFromTo with "
                    "'standard' sampling. Consider using 'negative_edge' "
                    "sampling instead."
                )

    def forward(
        self,
        model: models.Model,
        global_relations: dict[str, relationdata.RelationData],
        batch_relations: dict[str, relations.Relations],
        batch: dict[str, torch.Tensor],
        device: torch.device,
    ) -> torch.Tensor:

        batch_rel = batch_relations[self.batch_relation_key]
        data_key = batch_rel.data_key

        embed = getattr(model, self.embedding_method)

        if self._use_from_to:
            loss = self.loss_function(
                batch[self.global_relation_key].to(device),
                batch_rel.compute_relations(
                    embed(batch[self.data_key].to(device))
                ).data,  # type: ignore
            )
        else:
            global_rel_sub = (
                global_relations[self.global_relation_key]
                .sub(batch["indices"])
                .to(device)
            )
            if self.normalize_sub:
                global_rel_sub = global_rel_sub / global_rel_sub.sum()
            loss = self.loss_function(
                global_rel_sub,
                batch_rel.compute_relations(
                    embed(batch[data_key].to(device))
                ).data,  # type: ignore
            )
        return loss


class ClassificationLoss(Loss):
    """A loss that compares predicted class labels against ground truth labels.

    This loss compares predicted class labels to ground truth labels in a batch
    using a specified loss function (cross-entropy by default). Class labels
    are predicted by applying the model's :meth:`classify` method to the
    specified data entry of the input batch.

    Args:
        loss_function: The loss function to be applied.
        data_key: The key under which to find the data in the input batch.
        label_key: The key under which ground truth labels are stored in the
            input batch.
        classification_method: The model method to be used for classifying the
            batch of input data.
        name: Name of the loss (used by logging functions).
    """

    _prefix = "class_loss"

    def __init__(
        self,
        loss_function: BinaryTensorFun = torch.nn.CrossEntropyLoss(),
        data_key: str = "main",
        label_key: str = "labels",
        classification_method: str = "classify",
        name: Optional[str] = None,
    ):
        super().__init__(name)

        self.loss_function = loss_function
        self.data_key = data_key
        self.label_key = label_key
        self.classification_method = classification_method

    def forward(
        self,
        model: models.Model,
        global_relations: dict[str, relationdata.RelationData],
        batch_relations: dict[str, relations.Relations],
        batch: dict[str, torch.Tensor],
        device: torch.device,
    ) -> torch.Tensor:

        classify = getattr(model, self.classification_method)

        return self.loss_function(
            classify(batch[self.data_key].to(device)),
            batch[self.label_key].to(device),
        )


class PositionLoss(Loss):
    """A loss that compares embedding coordinates to given positions.

    This loss compares embedding coordiantes to given ground-truth coordinates
    in a batch using a specified loss function (mean-square-error by default).
    Embedding positions are computed by applying the model's :meth:`embed`
    method to the specified data entry of the input batch.

    Args:
        loss_function: The loss function to be applied.
        data_key: The key under which to find the data in the input batch.
        position_key: The key under which the ground truth positions are stored
            in the input batch.
        embedding_method: The model method to be used for embedding the batch
            of input data.
        name: Name of the loss (used by logging functions).
    """

    _prefix = "pos_loss"

    def __init__(
        self,
        loss_function: BinaryTensorFun = torch.nn.MSELoss(),
        data_key: str = "main",
        position_key: str = "pos",
        embedding_method: str = "embed",
        name: Optional[str] = None,
    ):
        super().__init__(name)

        self.loss_function = loss_function
        self.data_key = data_key
        self.position_key = position_key
        self.embedding_method = embedding_method

    def forward(
        self,
        model: models.Model,
        global_relations: dict[str, relationdata.RelationData],
        batch_relations: dict[str, relations.Relations],
        batch: dict[str, torch.Tensor],
        device: torch.device,
    ) -> torch.Tensor:

        embed = getattr(model, self.embedding_method)

        return self.loss_function(
            embed(batch[self.data_key].to(device)),
            batch[self.position_key].to(device),
        )


class ReconstructionLoss(Loss):
    """A simple reconstruction loss for auto-encoding data.

    This loss compares reconstructed data to input data in a batch using a
    specified loss function (mean-square-error by default). Reconstructed
    data is computed by applying the model's :meth:`decode` and :meth:`encode`
    methods subsequently to the specified data entry of the input batch.

    Args:
        loss_function: The loss function to be applied.
        data_key: The key under which to find the data in the input batch.
        encoding_method: The model method to be used for encoding the batch of
            input data.
        decoding_method: The model method to be used for decoding the encoded
            batch of input data.
        name: Name of the loss (used by logging functions).
    """

    _prefix = "recon_loss"

    def __init__(
        self,
        loss_function: BinaryTensorFun = torch.nn.MSELoss(),
        data_key: str = "main",
        encoding_method: str = "encode",
        decoding_method: str = "decode",
        name: Optional[str] = None,
    ):
        super().__init__(name)

        self.loss_function = loss_function
        self.data_key = data_key
        self.encoding_method = encoding_method
        self.decoding_method = decoding_method

    def forward(
        self,
        model: models.Model,
        global_relations: dict[str, relationdata.RelationData],
        batch_relations: dict[str, relations.Relations],
        batch: dict[str, torch.Tensor],
        device: torch.device,
    ) -> torch.Tensor:

        data = batch[self.data_key].to(device)
        encode = getattr(model, self.encoding_method)
        decode = getattr(model, self.decoding_method)

        return self.loss_function(data, decode(encode(data)))


class CompoundLoss(Loss):
    """A weighted sum of multiple losses.

    Args:
        losses: A list of :class:`paradime.loss.Loss` instances to be
            summed.
        weights: A list of weights to multiply the losses with. Must be of the
            same length as the list of losses. If no weights are specified,
            all losses are weighted equally.
        name: Name of the loss (used by logging functions).
    """

    _prefix = "comp_loss"

    def __init__(
        self,
        losses: list[Loss],
        weights: Union[np.ndarray, torch.Tensor, list[float], None] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name)

        self.losses = losses

        if weights is None:
            weights = torch.ones(len(losses))
        elif len(weights) != len(self.losses):
            raise ValueError("Size mismatch between losses and weights.")

        self.weights = utils.convert.to_torch(weights)

    def checkpoint(self) -> None:
        super().checkpoint()
        for l in self.losses:
            l.checkpoint()

    def detailed_history(self) -> torch.Tensor:
        """Returns a detailed history of the compound loss.

        Returns:
            A PyTorch tensor with the history of each loss component multiplied
            by its weight.
        """
        histories = torch.tensor([loss.history for loss in self.losses])
        return self.weights[:, None] * histories

    def _check_sampling_and_relations(
        self,
        sampling: Literal["standard", "negative_edge"],
        batch_relations: dict[str, relations.Relations],
    ) -> None:
        for loss in self.losses:
            if isinstance(loss, RelationLoss):
                loss._check_sampling_and_relations(sampling, batch_relations)

    def forward(
        self,
        model: models.Model,
        global_relations: dict[str, relationdata.RelationData],
        batch_relations: dict[str, relations.Relations],
        batch: dict[str, torch.Tensor],
        device: torch.device,
    ) -> torch.Tensor:

        total_loss = torch.tensor(0.0).to(device)

        for loss, w in zip(self.losses, self.weights.to(device)):
            loss_val = loss(
                model, global_relations, batch_relations, batch, device
            )
            total_loss += w * loss_val

        return total_loss


def kullback_leibler_div(
    p: torch.Tensor, q: torch.Tensor, epsilon: float = 1.0e-7
) -> torch.Tensor:
    """Kullback-Leibler divergence.

    To be used as a loss function in the :class:`paradime.loss.RelationLoss`
    of a parametric DR routine.

    Args:
        p: Input tensor containing (a batch of) probabilities.
        q: Input tensor containing (a batch of) probabilities.
        epsilon: Small constant used to avoid numerical errors caused by
            near-zero probability values.

    Returns:
        The Kullback-Leibler divergence of the two input tensors, divided by
        the number of items in the batch.
    """
    eps = torch.tensor(epsilon, dtype=p.dtype)
    kl_matr = torch.mul(p, torch.log(p + eps) - torch.log(q + eps))
    kl_matr.fill_diagonal_(0.0)
    return torch.sum(kl_matr) / len(p)


def cross_entropy_loss(
    p: torch.Tensor, q: torch.Tensor, epsilon: float = 1.0e-7
) -> torch.Tensor:
    """Cross-entropy loss as used by UMAP.

    To be used as a loss function in the :class:`paradime.loss.RelationLoss`
    of a parametric DR routine.

    Args:
        p: Input tensor containing (a batch of) probabilities.
        q: Input tensor containing (a batch of) probabilities.
        epsilon: Small constant used to avoid numerical errors caused by
            near-zero probability values.

    Returns:
        The cross-entropy loss of the two input tensors, divided by the number
        items in the batch.
    """
    attraction = -1.0 * p * torch.log(torch.clamp(q, min=epsilon, max=1.0))
    repulsion = (
        -1.0 * (1.0 - p) * torch.log(torch.clamp(1 - q, min=epsilon, max=1.0))
    )
    loss = attraction + repulsion
    return torch.sum(loss) / len(p)



"""Parsing utilities for ParaDime.

The :mod:`paradime.utils.parsing` module implements functions to validate
specifications, parse them, and prepare them for later use.
"""

from typing import Any, Union

from cerberus import Validator
from ruamel.yaml import YAML

from paradime.exceptions import SpecificationError

schema = {
    "derived data": {
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "name": {"type": "string", "required": True},
                "data func": {
                    "type": "string",
                    "required": True,
                    "allowed": ["pca", "spectral"],
                },
                "keys": {
                    "type": "list",
                    "required": True,
                    "schema": {
                        "type": "list",
                        "items": [
                            {"type": "string", "allowed": ["data", "rels"]},
                            {"type": "string"},
                        ],
                    },
                },
                "options": {"type": "dict"},
            },
        },
    },
    "relations": {
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "name": {"type": "string", "required": True},
                "level": {
                    "type": "string",
                    "allowed": ["global", "batch"],
                    "required": True,
                },
                "type": {
                    "type": "string",
                    "required": True,
                    "allowed": [
                        "precomp",
                        "pdist",
                        "neighbor",
                        "pdistdiff",
                        "fromto",
                    ],
                },
                "field": {
                    "type": "string",
                },
                "options": {"type": "dict"},
                "transforms": {
                    "type": "list",
                    "required": False,
                    "schema": {
                        "type": "dict",
                        "schema": {
                            "type": {
                                "type": "string",
                                "required": True,
                                "allowed": [
                                    "symmetrize",
                                    "normalize",
                                    "normalize rows",
                                    "perplexity",
                                    "t-dist",
                                    "connect",
                                ],
                            },
                            "options": {"type": "dict"},
                        },
                    },
                },
            },
        },
    },
    "losses": {
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "name": {"type": "string", "required": True},
                "type": {
                    "type": "string",
                    "required": True,
                    "allowed": [
                        "relation",
                        "classification",
                        "reconstruction",
                        "position",
                    ],
                },
                "func": {
                    "type": "string",
                    "allowed": [
                        "mse",
                        "kl div",
                        "cross entropy",
                        "umap cross entropy",
                    ],
                },
                "keys": {
                    "type": "dict",
                    "schema": {
                        "data": {"type": "list", "schema": {"type": "string"}},
                        "rels": {"type": "list", "schema": {"type": "string"}},
                        "methods": {
                            "type": "list",
                            "schema": {"type": "string"},
                        },
                    },
                },
            },
        },
    },
    "training phases": {
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "epochs": {"type": "integer"},
                "sampling": {
                    "type": "dict",
                    "schema": {
                        "type": {
                            "type": "string",
                            "allowed": ["item", "edge"],
                        },
                        "options": {"type": "dict"},
                    },
                },
                "optimizer": {
                    "type": "dict",
                    "schema": {
                        "type": {
                            "type": "string",
                            "required": True,
                            "allowed": ["adam", "sgd"],
                        },
                        "options": {
                            "type": "dict",
                        },
                    },
                },
                "loss": {
                    "type": "dict",
                    "required": True,
                    "schema": {
                        "components": {
                            "type": "list",
                            "required": True,
                            "schema": {"type": "string"},
                        },
                        "weights": {
                            "type": "list",
                            "schema": {"type": "float"},
                        },
                    },
                },
            },
        },
    },
}

spec_validator = Validator(schema=schema)


def validate_spec(file_or_spec: Union[str, dict]) -> dict[str, Any]:
    """Validates a ParaDime specification.

    Args:
        file_or_spec: The specification, either as a dictionary or as a path
            to a YAML/JSON file.

    Returns:
        The validated specification as a dictionary.

    Raises:
        :class:`paradime.exceptions.SpecificationError`: If the validation of
            the specification failed.
    """

    yaml = YAML(typ="safe")

    if isinstance(file_or_spec, dict):
        spec = file_or_spec
    elif isinstance(file_or_spec, str):
        with open(file_or_spec) as f:
            spec = yaml.load(f)
    else:
        raise TypeError(
            "Expected specification dict or path to YAML/JSON file."
        )

    if spec_validator.validate(spec):
        return spec
    else:
        raise SpecificationError(f"{spec_validator.errors}")



from ._version import __version__

# from paradime.dr import ParametricDR, TrainingPhase, Dataset
# from paradime import loss
# from paradime import relations
# from paradime import transforms
# from paradime import routines



import sys

sys.path.append("../../")

import copy

import numpy as np
import torch
import sklearn.datasets
import sklearn.decomposition
from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns

import paradime.dr
import paradime.relations
import paradime.transforms
import paradime.loss
import paradime.utils

paradime.utils.seed.seed_all(42)

diabetes = sklearn.datasets.load_diabetes()
data = diabetes["data"]


def mse(a, b):
    return torch.sum((a - b) ** 2)


df = pd.DataFrame(columns=["Model", "Batch size", "Loss"])

hd_pdist = torch.nn.functional.pdist(torch.tensor(data))

batch_sizes = [10, 50, 100, 221, 442]

for name, model in zip(
    ["Linear (10, 2)", "Non-linear (10, 2)", "Non-linear (10, 5, 2)"],
    [
        torch.nn.Linear(10, 2, bias=False),
        torch.nn.Sequential(
            torch.nn.Linear(10, 2, bias=True),
            torch.nn.Softplus(),
        ),
        torch.nn.Sequential(
            torch.nn.Linear(10, 5, bias=True),
            torch.nn.Softplus(),
            torch.nn.Linear(5, 2, bias=True),
            torch.nn.Softplus(),
        ),
    ],
):
    for bs in batch_sizes:
        for _ in range(10):

            pd_mds = paradime.dr.ParametricDR(
                model=copy.deepcopy(model),
                global_relations=paradime.relations.PDist(
                    transform=[
                        paradime.transforms.ToSquareTensor(),
                        # paradime.transforms.Functional(lambda x: x.float()),
                    ]
                ),
                batch_relations=paradime.relations.DifferentiablePDist(
                    transform=paradime.transforms.ToSquareTensor()
                ),
                dataset=data,
                verbose=True,
            )
            pd_mds.add_training_phase(
                epochs=500,
                batch_size=bs,
                learning_rate=0.02,
                loss=paradime.loss.RelationLoss(
                    loss_function=mse,
                    embedding_method="forward",
                    normalize_sub=False,
                ),
                report_interval=75,
            )
            pd_mds.train()

            loss = mse(
                hd_pdist,
                torch.nn.functional.pdist(pd_mds.apply(data)),
            )
            df = pd.concat(
                [df, pd.DataFrame([[name, bs, loss]], columns=df.columns)],
                ignore_index=True,
            )


class DirectModel(torch.nn.Module):
    def __init__(self, size, dim):
        super().__init__()
        self.coords = torch.nn.Linear(size, dim, bias=False)

    def forward(self, X):
        pass

    def embed(self, indices):
        return self.coords.weight.data[indices]


class DirectLoss(paradime.loss.Loss):
    def forward(
        self,
        model,
        global_relations,
        batch_relations,
        batch,
        device,
    ) -> torch.Tensor:
        global_rel = global_relations["rel"].sub(batch["indices"])
        batch_rel = (
            batch_relations["rel"]
            .compute_relations(model.coords.weight[batch["indices"]])
            .data
        )

        return mse(global_rel, batch_rel)


for _ in range(10):
    for bs in batch_sizes:

        pd_mds = paradime.dr.ParametricDR(
            model=DirectModel(2, len(data)),
            global_relations=paradime.relations.PDist(
                transform=[
                    paradime.transforms.ToSquareTensor(),
                    # paradime.transforms.Functional(lambda x: x.float()),
                ]
            ),
            batch_relations=paradime.relations.DifferentiablePDist(
                transform=paradime.transforms.ToSquareTensor()
            ),
            dataset=data,
            verbose=True,
        )
        pd_mds.add_training_phase(
            epochs=500,
            batch_size=len(data) // 5,
            learning_rate=0.01,
            loss=DirectLoss(),
            report_interval=75,
        )
        pd_mds.train()

        loss = mse(
            hd_pdist,
            torch.nn.functional.pdist(pd_mds.model.coords.weight.data),
        )

        df = pd.concat(
            [df, pd.DataFrame([["Direct", bs, loss]], columns=df.columns)],
            ignore_index=True,
        )


from sklearn.manifold import MDS


for _ in range(10):
    mds = torch.tensor(MDS(n_init=1).fit_transform(data))

    loss = mse(hd_pdist, torch.nn.functional.pdist(mds))

    df = pd.concat(
        [
            df,
            pd.DataFrame(
                [["SMACOF", batch_sizes[-1], loss]], columns=df.columns
            ),
        ],
        ignore_index=True,
    )

df.to_csv("mds_runs.csv")



"""Default models for ParaDime.

The :mod:`paradime.models` module implements default models for several
applications, such as a simple
:class:`paradime.models.FullyConnectedEmbeddingModel`.
"""

import torch


class Model(torch.nn.Module):
    """A placeholder model that lists all methods used by the losses defined
    in :mod:`paradime.loss`.
    """

    def embed(self, X: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def encode(self, X: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def decode(self, X: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def classify(self, X: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()


class FullyConnectedEmbeddingModel(Model):
    """A fully connected network for embeddings.

    Args:
        in_dim: Input dimension.
        out_dim: Output dimensions.
        hidden_dims: List of hidden layer dimensions.
    """

    def __init__(self, in_dim: int, out_dim: int, hidden_dims: list[int]):
        super().__init__()

        self.layers = torch.nn.ModuleList()

        cur_dim = in_dim
        for hdim in hidden_dims:
            self.layers.append(torch.nn.Linear(cur_dim, hdim))
            cur_dim = hdim
        self.layers.append(torch.nn.Linear(cur_dim, out_dim))

        self.alpha = torch.nn.Parameter(torch.tensor(1.0))

    def forward(self, x):
        for layer in self.layers[:-1]:
            # x = torch.sigmoid(layer(x))
            x = torch.nn.Softplus()(layer(x))
        out = self.layers[-1](x)
        return out

    def embed(self, x):
        return self.forward(x)



"""Relation data containers for ParaDime.

The :mod:`paradime.relationdata` module implements container classes for
various formats of relation data. The relation data containers are used by the
different :class:`paradime.relations.Relations` (see :mod:`paradime.relations`)
and :class:`paradime.transforms.RelationTransform` (see
:mod:`paradime.transforms`).
"""

import itertools
from typing import Optional, Literal
import warnings

import numpy as np
import scipy.sparse
from scipy.spatial import distance
import torch

from paradime import utils
from paradime.types import IndexList, Rels


class RelationData(utils._ReprMixin):
    """Base class for storing relations between data points."""

    def __init__(self):
        self._data = None

    @property
    def data(self) -> Rels:
        return self._data

    @data.setter
    def data(self, relations: Rels) -> None:
        self._data = relations

    def sub(self, indices: IndexList) -> torch.Tensor:
        """Subsamples the relation matrix based on item indices.

        Intended to be used for batch-wise subsampling of global relations.

        Args:
            indices: A flat list of item indices.

        Returns:
            A square PyTorch tensor consisting of all relations
            between items with the given indices.
        """
        raise NotImplementedError(
            f"Matrix subsampling not implemented for {type(self).__name__}."
        )

    def to_flat_array(self) -> "FlatRelationArray":
        """Converts the relations to a
        :class:`paradime.relationdata.FlatRelationArray`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to flat array not "
            f"implemented for {type(self).__name__}."
        )

    def to_flat_tensor(self) -> "FlatRelationTensor":
        """Converts the relations to a
        :class:`paradime.relationdata.FlatRelationTensor`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to flat tensor not "
            f"implemented for {type(self).__name__}."
        )

    def to_square_array(self) -> "SquareRelationArray":
        """Converts the relations to a
        :class:`paradime.relationdata.SquareRelationArray`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to square array not "
            f"implemented for {type(self).__name__}."
        )

    def to_square_tensor(self) -> "SquareRelationTensor":
        """Converts the relations to a
        :class:`paradime.relationdata.SquareRelationTensor`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to square tensor not "
            f"implemented for {type(self).__name__}."
        )

    def to_triangular_array(self) -> "TriangularRelationArray":
        """Converts the relations to a
        :class:`paradime.relationdata.TriangularRelationArray`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to triangular array not "
            f"implemented for {type(self).__name__}."
        )

    def to_triangular_tensor(self) -> "TriangularRelationTensor":
        """Converts the relations to a
        :class:`paradime.relationdata.TriangularRelationTensor`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to triangular tensor not "
            f"implemented for {type(self).__name__}."
        )

    def to_neighbor_tuple(self) -> "NeighborRelationTuple":
        """Converts the relations to a
        :class:`paradime.relationdata.NeighborRelationTuple`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to neighbor relation tuple not "
            f"implemented for {type(self).__name__}."
        )

    def to_sparse_array(self) -> "SparseRelationArray":
        """Converts the relations to a
        :class:`paradime.relationdata.SparseRelationArray`.

        Returns:
            The converted relations.
        """
        raise NotImplementedError(
            "Conversion to sparse array not "
            f"implemented for {type(self).__name__}."
        )


def relation_factory(
    relations: Rels,
    force_flat: bool = False,
) -> RelationData:
    """Creates a :class:`paradime.relationdata.RelationData` object from a
    variety of input formats.

    Args:
        relations: The relations, specified either as a flat array or tensor,
            a square array or tensor, a vector-form (triangular) array or
            tensor, a sparse array, or a tuple (n, r), where n is an array of
            neighor indices for each data point and r is an array of relation
            values of the same shape.
        force_flat: If set true, disables the check for triangular arrays and
            tensors. Useful if flat relation data might have a length equal
            to a triangular number.

    Returns:
        A :class:`paradime.relationdata.RelationData` object with a subclass
        depending on the input format.
    """

    if _is_square_array(relations):
        rd = SquareRelationArray(relations)  # type: ignore
    elif _is_square_tensor(relations):
        rd = SquareRelationTensor(relations)  # type: ignore
    elif _is_square_sparse(relations):
        rd = SparseRelationArray(relations)  # type: ignore
    elif _is_triangular_array(relations) and not force_flat:
        rd = TriangularRelationArray(relations)  # type: ignore
    elif _is_triangular_tensor(relations) and not force_flat:
        rd = TriangularRelationTensor(relations)  # type: ignore
    elif _is_flat_array(relations):
        rd = FlatRelationArray(relations)  # type: ignore
    elif _is_flat_tensor(relations):
        rd = FlatRelationTensor(relations)  # type: ignore
    elif _is_array_tuple(relations):
        rd = NeighborRelationTuple(relations)  # type: ignore
    else:
        raise TypeError(f"Input type not supported by {RelationData.__name__}.")

    return rd


class FlatRelationArray(RelationData):
    """Relation data in the form of a flat array of individual relations.

    Args:
        relations: A flat Numpy array of relation values.

    Attributes:
        data: The raw relation data.
    """

    def __init__(self, relations: np.ndarray):

        if not _is_flat_array(relations):
            raise ValueError("Expected vector-form array.")

        self._data = relations

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, relations: np.ndarray) -> None:
        self._data = relations

    def to_flat_array(self) -> "FlatRelationArray":
        return self

    def to_flat_tensor(self) -> "FlatRelationTensor":
        return FlatRelationTensor(torch.tensor(self.data))


class FlatRelationTensor(RelationData):
    """Relation data in the form of a flat tensor of individual relations.

    Args:
        relations: A flat PyTorch tensor of relation values.

    Attributes:
        data: The raw relation data.
    """

    def __init__(self, relations: torch.Tensor):

        if not _is_flat_tensor(relations):
            raise ValueError("Expected vector-form tensor.")

        self._data = relations

    @property
    def data(self) -> torch.Tensor:
        return self._data

    @data.setter
    def data(self, relations: torch.Tensor) -> None:
        self._data = relations

    def to_flat_array(self) -> "FlatRelationArray":
        return FlatRelationArray(self.data.detach().numpy())

    def to_flat_tensor(self) -> "FlatRelationTensor":
        return self


class SquareRelationArray(RelationData):
    """Relation data in the form of a square array.

    Args:
        relations: A square Numpy array of relation values.

    Attributes:
        data: The raw relation data.
    """

    def __init__(self, relations: np.ndarray):

        if not _is_square_array(relations):
            raise ValueError("Expected square array.")

        self._data = relations

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, relations: np.ndarray) -> None:
        self._data = relations

    def sub(self, indices: IndexList) -> torch.Tensor:
        dim = len(indices)
        if isinstance(indices, torch.Tensor):
            indices = indices.detach().cpu().numpy()
        indices = np.array(np.meshgrid(indices, indices)).T.reshape(-1, 2).T
        return torch.tensor(
            np.reshape(self.data[indices[0], indices[1]], (dim, dim))
        )

    def to_square_array(self) -> "SquareRelationArray":
        return self

    def to_square_tensor(self) -> "SquareRelationTensor":
        return SquareRelationTensor(torch.tensor(self.data))

    def to_triangular_array(self) -> "TriangularRelationArray":
        return TriangularRelationArray(distance.squareform(self.data))

    def to_triangular_tensor(self) -> "TriangularRelationTensor":
        return TriangularRelationTensor(
            torch.tensor(distance.squareform(self.data))
        )

    def to_sparse_array(self) -> "SparseRelationArray":
        return SparseRelationArray(scipy.sparse.csr_matrix(self.data))

    def to_neighbor_tuple(self) -> "NeighborRelationTuple":
        return NeighborRelationTuple(
            (
                np.tile(np.arange(self.data.shape[0]), (self.data.shape[0], 1)),
                self.data,
            )
        )


class SquareRelationTensor(RelationData):
    """Relation data in the form of a square tensor.

    Args:
        relations: A square PyTorch tensor of relation values.

    Attributes:
        data: The raw relation data.
    """

    def __init__(self, relations: torch.Tensor):

        if not _is_square_tensor(relations):
            raise ValueError("Expected square tensor.")

        self._data = relations

    @property
    def data(self) -> torch.Tensor:
        return self._data

    @data.setter
    def data(self, relations: torch.Tensor) -> None:
        self._data = relations

    def sub(self, indices: IndexList) -> torch.Tensor:
        dim = len(indices)
        if isinstance(indices, torch.Tensor):
            indices = indices.detach().cpu().numpy()
        indices = np.array(np.meshgrid(indices, indices)).T.reshape(-1, 2).T
        return self.data[indices[0], indices[1]].reshape(dim, dim)

    def to_square_array(self) -> "SquareRelationArray":
        return SquareRelationArray(self.data.detach().cpu().numpy())

    def to_square_tensor(self) -> "SquareRelationTensor":
        return self

    def to_triangular_array(self) -> "TriangularRelationArray":
        return TriangularRelationArray(
            distance.squareform(self.data.detach().numpy())
        )

    def to_triangular_tensor(self) -> "TriangularRelationTensor":
        i, j = torch.triu_indices(
            self.data.shape[0], self.data.shape[1], offset=1
        )
        return TriangularRelationTensor(self.data[i, j])

    def to_sparse_array(self) -> "SparseRelationArray":
        return SparseRelationArray(
            scipy.sparse.csr_matrix(self.data.detach().numpy())
        )

    def to_neighbor_tuple(self) -> "NeighborRelationTuple":
        rels = self.data.detach().numpy()
        return NeighborRelationTuple(
            (np.tile(np.arange(rels.shape[0]), (rels.shape[0], 1)), rels)
        )


class TriangularRelationArray(RelationData):
    """Relation data in 'triangular' vector-form.

    Args:
        relations: A Numpy array of relation values, as accepted by
            :func:`scipy.spatial.distance.squareform`.

    Attributes:
        data: The raw relation data.
    """

    def __init__(self, relations: np.ndarray):

        if not _is_triangular_array(relations):
            raise ValueError("Expected vector-form relation array.")

        self._data = relations

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, relations: np.ndarray) -> None:
        self._data = relations

    def sub(self, indices: IndexList) -> torch.Tensor:
        dim = utils.convert.triu_to_square_dim(len(self.data))
        combos = itertools.combinations(indices, 2)
        return torch.tensor(
            distance.squareform(
                np.array(
                    [
                        self.data[utils.convert.rowcol_to_triu_index(i, j, dim)]
                        for i, j in combos
                    ]
                )
            )
        )

    def to_square_array(self) -> "SquareRelationArray":
        return SquareRelationArray(distance.squareform(self.data))

    def to_square_tensor(self) -> "SquareRelationTensor":
        return SquareRelationTensor(
            torch.tensor(distance.squareform(self.data))
        )

    def to_triangular_array(self) -> "TriangularRelationArray":
        return self

    def to_triangular_tensor(self) -> "TriangularRelationTensor":
        return TriangularRelationTensor(torch.tensor(self.data))

    def to_sparse_array(self) -> "SparseRelationArray":
        return SparseRelationArray(
            scipy.sparse.csr_matrix(distance.squareform(self.data))
        )

    def to_neighbor_tuple(self) -> "NeighborRelationTuple":
        return self.to_square_array().to_neighbor_tuple()


class TriangularRelationTensor(RelationData):
    """Relation data in 'triangular' vector-form.

    Args:
        relations: A PyTorch tensor of relation values, with a shape as
            accepted by :func:`scipy.spatial.distance.squareform`.

    Attributes:
        data: The raw relation data.
    """

    def __init__(self, relations: torch.Tensor):

        if not _is_triangular_tensor(relations):
            raise ValueError("Expected vector-form relation tensor.")

        self._data = relations

    @property
    def data(self) -> torch.Tensor:
        return self._data

    @data.setter
    def data(self, relations: torch.Tensor) -> None:
        self._data = relations

    def sub(self, indices: IndexList) -> torch.Tensor:
        dim = utils.convert.triu_to_square_dim(len(self.data))
        combos = itertools.combinations(indices, 2)
        return _tensor_squareform(
            torch.tensor(
                [
                    self.data[utils.convert.rowcol_to_triu_index(i, j, dim)]
                    for i, j in combos
                ]
            )
        )

    def to_square_array(self) -> "SquareRelationArray":
        return SquareRelationArray(
            distance.squareform(self.data.detach().numpy())
        )

    def to_square_tensor(self) -> "SquareRelationTensor":
        # get dimensions of square matrix
        d = utils.convert.triu_to_square_dim(len(self.data))
        matrix = torch.zeros(
            (d, d), dtype=self.data.dtype, device=self.data.device
        )
        a, b = torch.triu_indices(d, d, offset=1)
        matrix[[a, b]] = self.data
        matrix = matrix + matrix.T
        return SquareRelationTensor(matrix)

    def to_triangular_array(self) -> "TriangularRelationArray":
        return TriangularRelationArray(self.data.detach().numpy())

    def to_triangular_tensor(self) -> "TriangularRelationTensor":
        return self

    def to_sparse_array(self) -> "SparseRelationArray":
        return self.to_triangular_array().to_sparse_array()

    def to_neighbor_tuple(self) -> "NeighborRelationTuple":
        return self.to_square_array().to_neighbor_tuple()


class SparseRelationArray(RelationData):
    """Relation data in sparse array form.

    Args:
        relations: A square, sparse Scipy array of relation values.

    Attributes:
        data: The raw relation data.
    """

    def __init__(self, relations: scipy.sparse.spmatrix):

        if not _is_square_sparse(relations):
            raise ValueError("Expected vector-form relation tensor.")

        self._data = relations

    @property
    def data(self) -> scipy.sparse.spmatrix:
        return self._data

    @data.setter
    def data(self, relations: scipy.sparse.spmatrix) -> None:
        self._data = relations

    def sub(self, indices: IndexList) -> torch.Tensor:
        dim = len(indices)
        if isinstance(indices, torch.Tensor):
            indices = indices.detach().cpu().numpy()
        indices = np.array(np.meshgrid(indices, indices)).T.reshape(-1, 2).T
        return torch.tensor(self.data[indices[0], indices[1]]).reshape(dim, dim)

    def to_square_array(self) -> "SquareRelationArray":
        return SquareRelationArray(self.data.toarray())

    def to_square_tensor(self) -> "SquareRelationTensor":
        return SquareRelationTensor(torch.tensor(self.data.toarray()))

    def to_triangular_array(self) -> "TriangularRelationArray":
        indices = np.triu_indices(self.data.shape[0], k=1)
        return TriangularRelationArray(self.data[indices].A[0])

    def to_triangular_tensor(self) -> "TriangularRelationTensor":
        indices = np.triu_indices(self.data.shape[0], k=1)
        return TriangularRelationTensor(torch.tensor(self.data[indices].A[0]))

    def to_sparse_array(self) -> "SparseRelationArray":
        return self

    def to_neighbor_tuple(self) -> "NeighborRelationTuple":
        return self.to_square_array().to_neighbor_tuple()


class NeighborRelationTuple(RelationData):
    """Relation data in neighborhood tuple form.

    Args:
        relations: A tuple (n, r) of relation data, where n is an array of
            neighor indices for each data point and r is an array of relation
            values. Both arrays must be of shape (num_points, num_neighbors).
        sort: Sorting option. If None is passed (default), values are kept
            as is. Otherwise, values for each item are sorted either in
            ``'ascending'`` or ``'descending'`` order.

    Attributes:
        data: The raw relation data.
    """

    def __init__(
        self,
        relations: tuple[np.ndarray, np.ndarray],
        sort: Optional[Literal["ascending", "descending"]] = None,
    ):

        if not _is_array_tuple(relations):
            raise ValueError("Expected tuple of arrays (neighbors, relations).")

        neighbors, rels = relations

        if sort is None:
            pass
        elif sort == "ascending":
            indices = rels.argsort()
            neighbors = np.take_along_axis(neighbors, indices, axis=1)
            rels = np.take_along_axis(rels, indices, axis=1)
        elif sort == "descending":
            indices = rels.argsort()[:, ::-1]
            neighbors = np.take_along_axis(neighbors, indices, axis=1)
            rels = np.take_along_axis(rels, indices, axis=1)
        else:
            raise ValueError(
                f"Unknown sorting option {sort}. Only None, 'ascending' "
                "or 'descending are supported."
            )

        self._data: tuple[np.ndarray, np.ndarray] = (neighbors, rels)

    @property
    def data(self) -> tuple[np.ndarray, np.ndarray]:
        return self._data

    @data.setter
    def data(self, relations: tuple[np.ndarray, np.ndarray]) -> None:
        self._data = relations

    def _where_self_relations(self) -> np.ndarray:
        relations = self.data[0]
        return relations == np.arange(len(relations))[:, None]

    def _has_all_self_relations(self) -> bool:
        return bool(np.all(np.any(self._where_self_relations(), axis=1)))

    def _remove_self_relations(self) -> None:
        if not self._has_all_self_relations():
            warnings.warn(
                "Neighbor relation tuple does not include all self-relations. "
                "Removal would lead to ragged array and is not performed."
            )
        else:
            mask = np.logical_not(self._where_self_relations())
            self.data = (
                np.stack([i[j] for i, j in zip(self.data[0], mask)]),
                np.stack([i[j] for i, j in zip(self.data[1], mask)]),
            )

    def sub(self, indices: IndexList) -> torch.Tensor:
        return self.to_sparse_array().sub(indices)

    def to_square_array(self) -> "SquareRelationArray":
        return self.to_sparse_array().to_square_array()

    def to_square_tensor(self) -> "SquareRelationTensor":
        return self.to_sparse_array().to_square_tensor()

    def to_triangular_array(self) -> "TriangularRelationArray":
        return self.to_sparse_array().to_triangular_array()

    def to_triangular_tensor(self) -> "TriangularRelationTensor":
        return self.to_sparse_array().to_triangular_tensor()

    def to_sparse_array(self) -> "SparseRelationArray":
        neighbors, rels = self.data
        row_indices = np.repeat(
            np.arange(neighbors.shape[0]), neighbors.shape[1]
        )
        matrix = scipy.sparse.csr_matrix(
            (rels.ravel(), (row_indices, neighbors.ravel()))
        )
        return SparseRelationArray(matrix)

    def to_neighbor_tuple(self) -> "NeighborRelationTuple":
        return self


def _is_square_array(rels: Rels) -> bool:
    result = False
    if isinstance(rels, np.ndarray):
        if len(rels.shape) == 2 and rels.shape[0] == rels.shape[1]:
            result = True
    return result


def _is_square_tensor(rels: Rels) -> bool:
    result = False
    if isinstance(rels, torch.Tensor):
        if len(rels.shape) == 2 and rels.shape[0] == rels.shape[1]:
            result = True
    return result


def _is_triangular_array(rels: Rels) -> bool:
    result = False
    if isinstance(rels, np.ndarray):
        s = rels.shape
        if len(s) == 1:
            d = int(np.ceil(np.sqrt(s[0] * 2)))
            if d * (d - 1) == s[0] * 2:
                result = True
    return result


def _is_triangular_tensor(rels: Rels) -> bool:
    result = False
    if isinstance(rels, torch.Tensor):
        s = rels.shape
        if len(s) == 1:
            d = int(np.ceil(np.sqrt(s[0] * 2)))
            if d * (d - 1) == s[0] * 2:
                result = True
    return result


def _is_flat_array(rels: Rels) -> bool:
    result = False
    if isinstance(rels, np.ndarray):
        if len(rels.shape) == 1:
            result = True
    return result


def _is_flat_tensor(rels: Rels) -> bool:
    result = False
    if isinstance(rels, torch.Tensor):
        if len(rels.shape) == 1:
            result = True
    return result


def _is_square_sparse(rels: Rels) -> bool:
    result = False
    if isinstance(rels, scipy.sparse.spmatrix):
        if len(rels.shape) == 2 and rels.shape[0] == rels.shape[1]:
            result = True
    return result


def _is_array_tuple(rels: Rels) -> bool:
    result = False
    if isinstance(rels, tuple) and len(rels) == 2:
        if len(rels[0].shape) == 2 and rels[0].shape == rels[1].shape:
            if rels[0].max() < len(rels[0]):
                result = True
    return result


def _tensor_squareform(t: torch.Tensor) -> torch.Tensor:
    d = utils.convert.triu_to_square_dim(len(t))
    matrix = torch.zeros((d, d), dtype=t.dtype, device=t.device)
    a, b = torch.triu_indices(d, d, offset=1)
    matrix[[a, b]] = t
    return matrix + matrix.T



"""Predefined ParaDime routines for existing DR techniques.

The :mod:`paradime.routines` module implements parametric versions of existing
dimensionality reduction techniques using the :class:`paradime.dr.ParametricDR`
interface.
"""

from typing import Optional, Union

import torch
import sklearn.decomposition
import sklearn.manifold

from paradime import dr
from paradime import loss as pdloss
from paradime import models
from paradime import relations
from paradime import transforms


class ParametricTSNE(dr.ParametricDR):
    """A parametric version of t-SNE.

    This class provides a high-level interface for a
    :class:`paradime.paradime.ParametricDR` routine with the following
    specifications:

    * The global relations are :class:`paradime.relations.NeighborBasedPDist`,
      transformed with a :class:`paradime.transforms.PerplexityBasedRescale`
      followed by :class:`paradime.tranforms.Symmetrize`.
    * The batch relations are :class:`paradime.relations.DifferentiablePDist`,
      transformed with a :class:`paradime.relations.StudentTTransform`
      followed by :class:`paradime.transform.Normalize`.
    * The first (optional) training phase intializes the model to approximate
      PCA (see ``intialization`` below).
    * The second training phase uses the Kullback-Leibler divergence to compare
      the relations.

    Args:
        perplexity: The desired perplexity, which can be understood as
            a smooth measure of nearest neighbors used to determine
            high-dimensional relations between data points.
        alpha: Degrees of freedom of the Student's t-disitribution used to
            calculate low-dimensional relations between data points.
        model: The model used to embed the high dimensional data.
        in_dim: The numer of dimensions of the input data, used to construct a
            default model in case none is specified. If a dataset is specified
            at instantiation, the correct value for this parameter will be
            inferred from the data dimensions.
        out_dim: The number of output dimensions (i.e., the dimensionality of
            the embedding).
        hidden_dims: Dimensions of hidden layers for the default fully
            connected model that is created if no model is specified.
        initialization: How to pretrain the model to mimic initialization of
            low-dimensional positions. By default (``'pca'``) the model is
            pretrained to output an approximation of PCA before beginning the
            main training phase.
        epochs: The number of epochs in the main training phase.
        init_epochs: The number of epochs in the pretraining (initialization).
            phase.
        batch_size: The number of items in a batch during the main training
            phase.
        init_batch_size: The number of items in a batch during the pretraining
            (initialization).
        learning_rate: The learning rate during the main training phase.
        init_learning_reate: The learning rate during the pretraining
            (initialization).
        data_key: The key under which the data can be found in the dataset.
        dataset: The dataset on which to perform the training. Datasets can be
            registerd after instantiation using the :meth:`register_dataset`
            class method.
        use_cuda: Whether or not to use the GPU for training.
        verbosity: Verbosity flag.
    """

    def __init__(
        self,
        perplexity: float = 30.0,
        alpha: float = 1.0,
        model: Optional[models.Model] = None,
        in_dim: Optional[int] = None,
        out_dim: int = 2,
        hidden_dims: list[int] = [100, 50],
        initialization: Optional[str] = "pca",
        epochs: int = 30,
        init_epochs: int = 10,
        batch_size: int = 500,
        init_batch_size: Optional[int] = None,
        learning_rate: float = 0.01,
        init_learning_rate: Optional[float] = None,
        data_key: str = "main",
        use_cuda: bool = False,
        verbose: bool = False,
    ):
        self.out_dim = out_dim
        self.initialization = initialization
        self.epochs = epochs
        self.init_epochs = init_epochs
        self.batch_size = batch_size
        if init_batch_size is None:
            self.init_batch_size = self.batch_size
        else:
            self.init_batch_size = init_batch_size
        self.learning_rate = learning_rate
        if init_learning_rate is None:
            self.init_learning_rate = self.learning_rate
        else:
            self.init_learning_rate = init_learning_rate
        self.data_key = data_key

        global_rel = relations.NeighborBasedPDist(
            transform=[
                transforms.PerplexityBasedRescale(perplexity=perplexity),
                transforms.Symmetrize(),
                transforms.Normalize(),
            ]
        )
        batch_rel = relations.DifferentiablePDist(
            transform=[
                transforms.StudentTTransform(alpha=alpha),
                transforms.Normalize(),
                transforms.ToSquareTensor(),
            ]
        )

        derived_data: dict[str, dr.DerivedData] = {}
        if self.initialization == "pca":
            derived_data["pca"] = dr.DerivedData(
                dr._pca,
                type_key_tuples=[("data", self.data_key)],
                n_components=self.out_dim,
            )

        losses = {
            "init": pdloss.PositionLoss(position_key="pca"),
            "embedding": pdloss.RelationLoss(
                loss_function=pdloss.kullback_leibler_div
            ),
        }

        init_phase = dr.TrainingPhase(
            name="pca_init",
            loss_keys=["init"],
            batch_size=self.init_batch_size,
            epochs=self.init_epochs,
            learning_rate=self.init_learning_rate,
        )

        main_phase = dr.TrainingPhase(
            name="embedding",
            loss_keys=["embedding"],
            batch_size=self.batch_size,
            epochs=self.epochs,
            learning_rate=self.learning_rate,
        )

        super().__init__(
            model=model,
            in_dim=in_dim,
            out_dim=out_dim,
            hidden_dims=hidden_dims,
            derived_data=derived_data,
            global_relations=global_rel,
            batch_relations=batch_rel,
            losses=losses,
            training_phases=[init_phase, main_phase],
            use_cuda=use_cuda,
            verbose=verbose,
        )


class ParametricUMAP(dr.ParametricDR):
    """A parametric version of UMAP.

    This class provides a high-level interface for a
    :class:`paradime.paradime.ParametricDR` routine with the following
    specifications:

    * The global relations are :class:`paradime.relations.NeighborBasedPDist`,
      transformed with a :class:`paradime.transforms.ConnectivityBasedRescale`
      followed by :class:`paradime.tranforms.Symmetrize` with product
      subtraction.
    * The batch relations are :class:`paradime.relations.DistsFromTo` (since
      negative edge sampling is used), transformed with a
      :class:`paradime.relations.ModifiedCauchyTransform`.
    * The first (optional) training phase intializes the model to approximate
      a spectral embedding based on the global relations (see ``intialization``
      below).
    * The second training phase uses corss-entropy to compare the relations.
      This phase uses negative edge sampling.

    Args:
        n_neighbors: The desired number of neighbors used for computing the
            high-dimensional pairwise relations.
        min_dist: Effective minimum distance of points in the embedding.
        spread: Effective scale of the points in the embedding.
        a: Parameter to define the modified Cauchy distribution used to compute
            low-dimensional relations.
        b: Parameter to define the modified Cauchy distribution used to compute
            low-dimensional relations.
        model: The model used to embed the high dimensional data.
        in_dim: The numer of dimensions of the input data, used to construct a
            default model in case none is specified. If a dataset is specified
            at instantiation, the correct value for this parameter will be
            inferred from the data dimensions.
        out_dim: The number of output dimensions (i.e., the dimensionality of
            the embedding).
        hidden_dims: Dimensions of hidden layers for the default fully
            connected model that is created if no model is specified.
        initialization: How to pretrain the model to mimic initialization of
            low-dimensional positions. By default (``'spectral'``) the model is
            pretrained to output an approximation of a soectral embedding based
            on the high-dimensional relations before beginning the main
            training phase.
        epochs: The number of epochs in the main training phase.
        init_epochs: The number of epochs in the pretraining (initialization).
            phase.
        batch_size: The number of items in a batch during the main training
            phase.
        init_batch_size: The number of items in a batch during the pretraining
            (initialization).
        learning_rate: The learning rate during the main training phase.
        init_learning_reate: The learning rate during the pretraining
            (initialization).
        data_key: The key under which the data can be found in the dataset.
        dataset: The dataset on which to perform the training. Datasets can be
            registerd after instantiation using the :meth:`register_dataset`
            class method.
        use_cuda: Whether or not to use the GPU for training.
        verbosity: Verbosity flag.
    """

    def __init__(
        self,
        n_neighbors: int = 30,
        min_dist: float = 0.01,
        spread: float = 1.0,
        a: Optional[float] = None,
        b: Optional[float] = None,
        model: Optional[models.Model] = None,
        in_dim: Optional[int] = None,
        out_dim: int = 2,
        hidden_dims: list[int] = [100, 50],
        initialization: Optional[str] = "spectral",
        epochs: int = 30,
        init_epochs: int = 5,
        batch_size: int = 10,
        negative_sampling_rate: int = 5,
        init_batch_size: int = 100,
        learning_rate: float = 0.005,
        init_learning_rate: float = 0.05,
        data_key: str = "main",
        dataset: Optional[Union[dr.Data, dr.Dataset]] = None,
        use_cuda: bool = False,
        verbose: bool = False,
    ):
        self.out_dim = out_dim
        self.initialization = initialization
        self.epochs = epochs
        self.init_epochs = init_epochs
        self.batch_size = batch_size
        self.negative_sampling_rate = negative_sampling_rate
        self.init_batch_size = init_batch_size
        self.learning_rate = learning_rate
        self.init_learning_rate = init_learning_rate
        self.data_key = data_key

        global_rel = relations.NeighborBasedPDist(
            transform=[
                transforms.ConnectivityBasedRescale(n_neighbors=n_neighbors),
                transforms.Symmetrize(subtract_product=True),
            ]
        )
        batch_rel = relations.DistsFromTo(
            transform=[
                transforms.ModifiedCauchyTransform(
                    min_dist=min_dist,
                    spread=spread,
                    a=a,
                    b=b,
                )
            ]
        )

        derived_data: dict[str, dr.DerivedData] = {}
        if self.initialization == "spectral":
            derived_data["spectral"] = dr.DerivedData(
                dr._spectral,
                type_key_tuples=[("rels", "rel")],
                n_components=self.out_dim,
            )

        losses = {
            "spectral_init": pdloss.PositionLoss(position_key="spectral"),
            "embedding": pdloss.RelationLoss(
                loss_function=pdloss.cross_entropy_loss
            ),
        }

        init_phase = dr.TrainingPhase(
            name="spectral_init",
            loss_keys=["spectral_init"],
            batch_size=self.init_batch_size,
            epochs=self.init_epochs,
            learning_rate=self.init_learning_rate,
        )

        main_phase = dr.TrainingPhase(
            name="embedding",
            loss_keys=["embedding"],
            sampling="negative_edge",
            neg_sampling_rate=self.negative_sampling_rate,
            batch_size=self.batch_size,
            epochs=self.epochs,
            learning_rate=self.learning_rate,
        )

        super().__init__(
            model=model,
            in_dim=in_dim,
            out_dim=out_dim,
            hidden_dims=hidden_dims,
            derived_data=derived_data,
            global_relations=global_rel,
            batch_relations=batch_rel,
            losses=losses,
            training_phases=[init_phase, main_phase],
            use_cuda=use_cuda,
            verbose=verbose,
        )



"""Custom exceptions for ParaDime.
"""


class NotTrainedError(Exception):
    """Method of DR instance called before training."""


class RelationsNotComputedError(Exception):
    """Attempted to access relation data before computation."""


class NoDatasetRegisteredError(Exception):
    """Attempted to access the dataset before registration."""


class LossNotDeterminedError(Exception):
    """Attempted to access loss before it was determined."""


class UnsupportedConfigurationError(Exception):
    """Combination of parameters and/or objects not supported."""


class SpecificationError(Exception):
    """Specification not supported."""


class CUDANotAvailableError(Exception):
    """CUDA not available."""



"""Loggin utility for ParaDime.

The :mod:`paradime.utils.logging` module implements logging functionality used
by verbose ParaDime routines.
"""

import logging
import sys

LOGGING_FORMAT = "%(asctime)s: %(message)s"
LOGGER_NAME = "paradime"

logger = logging.getLogger(LOGGER_NAME)
formatter = logging.Formatter(LOGGING_FORMAT)

logger.setLevel(logging.INFO)

ch = logging.StreamHandler(sys.stdout)
ch.setFormatter(formatter)

logger.addHandler(ch)


def log(message: str) -> None:
    """Calls the ParaDime logger to print a timestamp and a message.

    Args:
        message: The message string to print.
    """
    logger = logging.getLogger(LOGGER_NAME)
    logger.info(message)


def set_logfile(
    filename: str,
    mode: str = "a",
    disable_stdout: bool = False,
    disable_other_files: bool = False,
) -> None:
    """Configure the ParaDime logger to write its output to a file.

    Args:
        filename: The path to the log file.
        mode: The mode to open the file.
        disable_stdout: Whether or not to disbale logging to stdout.
        disable_other_files: Whether or not to remove other file handlers from
            the ParaDime logger.
    """
    logger = logging.getLogger(LOGGER_NAME)

    if disable_stdout:
        logger.removeHandler(ch)
    if disable_other_files:
        for h in logger.handlers:
            if isinstance(h, logging.FileHandler):
                logger.removeHandler(h)
    fh = logging.FileHandler(filename=filename, mode=mode, encoding="utf-8")
    fh.setFormatter(formatter)
    logger.addHandler(fh)



"""Random seeding for ParaDime.

The :mod:`paradime.utils.seed` subpackage implements a function to seed all
random number generators potentially involved in a ParaDime routine.
"""

import os
import random

import numpy as np
from packaging import version
import torch


def seed_all(seed: int) -> torch.Generator:
    """Sets several seeds to maximize reproducibility.

    For infos on reproducibility in PyTorch, see
    https://pytorch.org/docs/stable/notes/randomness.html.

    Args:
        seed: The integer to use as a seed.

    Returns:
        The :class:`torch.Generator` instance returned by
        :func:`torch.manual_seed`.
    """
    os.environ["PYTHONHASHSEED"] = str(seed)
    if torch.cuda.is_available():
        if version.parse(torch.version.cuda) >= version.parse("10.2"):
            os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
    random.seed(seed)
    np.random.seed(seed)
    gen = torch.manual_seed(seed)
    torch.use_deterministic_algorithms(True)
    return gen



"""Main module of ParaDime.

The :mod:`paradime.dr` module implements the main functionality of ParaDime.
This includes the :class:`paradime.dr.ParametricDR` class, as well as
:class:`paradime.dr.Dataset` and :class:`paradime.dr.TrainingPhase`.
"""

import copy
from typing import Callable, Literal, Mapping, Optional, Type, TypeVar, Union

import numpy as np
import torch
import sklearn.decomposition
import sklearn.manifold

from paradime import exceptions
from paradime import models
from paradime import relationdata
from paradime import relations
from paradime import transforms
from paradime import loss as pdloss
from paradime.types import TensorLike, TypeKeyTuples
from paradime import utils


class DerivedData:
    """A derived dataset entry to be computed later.

    Derived dataset entries can be used to set up rules for extending existing
    datasets later based on functions acting on other dataset entries or
    global relations.

    Args:
        func: The function to compute the derived data.
        type_key_tuples: A list of (type, key) tuples, where the types can be
            ``'data'`` or ``'rels'``, and the keys are used to access the
            respective entries.
    """

    def __init__(
        self,
        func: Callable[..., TensorLike],
        type_key_tuples: TypeKeyTuples = [("data", "main")],
        **kwargs,
    ):

        self.func = func
        self.requires_relations = False
        self.type_key_tuples = type_key_tuples
        self.kwargs = kwargs

        types, keys = list(zip(*type_key_tuples))

        for t in types:
            if t == "data":
                pass
            elif t == "rels":
                self.requires_relations = True
            else:
                raise ValueError(
                    "Expected 'data' or 'rels' as argument type "
                    f"for derived entry. Found '{t}' instead."
                )


Data = Union[
    np.ndarray,
    torch.Tensor,
    Mapping[str, Union[np.ndarray, torch.Tensor]],
]


class Dataset(torch.utils.data.Dataset, utils._ReprMixin):
    """A dataset for dimensionality reduction.

    Constructs a PyTorch :class:torch.utils.data.Dataset from the given data
    in such a way that each item or batch of items is a dictionary with
    PyTorch tensors as values. If only a single numpy array or PyTorch tensor
    is passed, this data will be available under the ``'data'`` key of the
    dict. Alternatively, a dict of tensors and/or arrays can be passed, which
    allows additional data such as labels for supervised learning. By default,
    an entry for indices is added to the dict, if it is not yet included in the
    passed dict.

    Args:
        data: The data, passed either as a single numpy array or PyTorch
            tensor, or as a dictionary containing multiple arrays and/or
            tensors.
    """

    def __init__(self, data: Data):

        self.data: dict[str, torch.Tensor] = {}

        if isinstance(data, (np.ndarray, torch.Tensor)):
            data = {"main": data}
        elif not isinstance(data, dict):
            raise ValueError(
                "Expected numpy array, PyTorch tensor, or dict "
                f"instead of {type(data)}."
            )
        # else:
        #     if "data" not in data:
        #         raise AttributeError(
        #             "Dataset expects a dict with a 'data' entry."
        #         )

        for i, (k, val) in enumerate(data.items()):
            if not isinstance(val, (np.ndarray, torch.Tensor)):
                raise ValueError(
                    f"Value for key {k} is not a numpy array or "
                    "PyTorch tensor."
                )
            elif i == 0:
                self.num_items = len(val)
                self.data[k] = utils.convert.to_torch(val)
            elif len(val) != self.num_items:
                raise ValueError(
                    "Dataset dict must have values of equal length."
                )
            else:
                self.data[k] = utils.convert.to_torch(val)

        if "indices" not in self.data:
            self.data["indices"] = torch.arange(len(self))

    def __len__(self) -> int:
        return self.num_items

    def __getitem__(self, index) -> dict[str, torch.Tensor]:
        out = {}
        for k in self.data:
            out[k] = self.data[k][index]
        return out


class NegSampledEdgeDataset(torch.utils.data.Dataset):
    """A dataset that supports negative edge sampling.

    Constructs a PyTorch :class:`torch.utils.data.Dataset` suitable for
    negative sampling from a regular :class:Dataset. The passed relation
    data, along with the negative samplnig rate ``r``, is used to inform the
    negative sampling process. Each \"item\" ``i`` of the resulting dataset
    is essentially a small batch of items, including the item ``i`` of the
    original dataset, one of it's actual neighbors, and ``r`` random other
    items that are considered to not be neighbors of ``i``. Remaining data
    from the original dataset is collated using PyTorch's
    :func:`torch.utils.data.default_collate` method.

    Args:
        data: The data in the form of a ParaDime :class:`paradime.dr.Dataset`.
        relations: A :class:`paradime.relationdata.RelationData` object with
            the edge data used for negative edge sampling.
        neg_sampling_rate: The negative sampling rate.
    """

    def __init__(
        self,
        dataset: Dataset,
        relations: relationdata.RelationData,
        neg_sampling_rate: int = 5,
        data_key: str = "main",
    ):

        self.dataset = dataset
        self.p_ij = relations.to_sparse_array().data.tocoo()
        self.weights = self.p_ij.data
        self.neg_sampling_rate = neg_sampling_rate
        self.data_key = data_key

        if self.data_key not in self.dataset.data:
            raise KeyError(
                f"Dataset does not have an attribute '{self.data_key}'."
            )

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        # the following assumes that the input relations are symmetric.
        # if they are not, the rows and cols should be shuffled

        # make nsr + 1 copies of row index
        rows = torch.full(
            (self.neg_sampling_rate + 1,), self.p_ij.row[idx], dtype=torch.long
        )

        # pick nsr + 1 random col indices (negative samples)
        cols = torch.randint(
            self.p_ij.shape[0], (self.neg_sampling_rate + 1,), dtype=torch.long
        )
        # set only one to an actual neighbor
        cols[0] = self.p_ij.col[idx]

        # make simplified p_ij (0 or 1)
        p_simpl = torch.zeros(self.neg_sampling_rate + 1, dtype=torch.float32)
        p_simpl[0] = 1

        indices = torch.tensor(
            np.unique(np.concatenate((rows.numpy(), cols.numpy())))
        )

        edge_data = {"row": rows, "col": cols, "rel": p_simpl}

        edge_data["from_to_data"] = torch.stack(
            (
                self.dataset.data[self.data_key][rows],
                self.dataset.data[self.data_key][cols],
            )
        )

        remaining_data = torch.utils.data.default_collate(
            [self.dataset[i] for i in indices]
        )

        return {**remaining_data, **edge_data}


def _collate_edge_batch(
    raw_batch: list[dict[str, torch.Tensor]]
) -> dict[str, torch.Tensor]:

    indices, unique_ids = np.unique(
        torch.concat([i["indices"] for i in raw_batch]), return_index=True
    )

    collated_batch = {"indices": torch.tensor(indices)}

    for k in raw_batch[0]:
        if k in ["row", "col", "rel"]:
            collated_batch[k] = torch.concat([i[k] for i in raw_batch])
        elif k == "from_to_data":
            collated_batch[k] = torch.concat([i[k] for i in raw_batch], dim=1)
        else:
            collated_batch[k] = torch.concat([i[k] for i in raw_batch])[
                torch.tensor(unique_ids)
            ]

    return collated_batch


class TrainingPhase(utils._ReprMixin):
    """A collection of parameter settings for a single phase in the
    training of a :class:`paradime.dr.ParametricDR` instance.

    Args:
        name: The name of the training phase.
        epochs: The number of epochs to run in this phase. In standard
            item-based sampling, the model sees every item once per epoch
            In the case of negative edge sampling, this is not guaranteed, and
            an epoch instead comprises ``batches_per_epoch`` batches (see
            parameter description below).
        batch_size: The number of items/edges in a batch. In standard
            item-based sampling, a batch has this many items, and the edges
            used for batch relations are constructed from the items. In the
            case of negative edge sampling, this is the number of sampled
            *positive* edges. The total number of edges is higher by a factor
            of ``r + 1``, where ``r`` is the negative sampling rate. The same holds
            for the number of items (apart from possible duplicates, which can
            result from the edge sampling and are removed).
        batches_per_epoch: The number of batches per epoch. This parameter
            only has an effect for negative edge sampling, where the number
            of batches per epoch is not determined by the dataset size and the
            batch size. If this parameter is set to -1 (default), an epoch
            will comprise a number of batches that leads to a total number of
            sampled *items* roughly equal to the number of items in the
            dataset. If this parameter is set to an integer, an epoch will
            instead comprise that many batches.
        sampling: The sampling strategy, which can be either ``'standard'``
            (simple item-based sampling; default) or ``'negative_edge'``
            (negative edge sampling).
        edge_rel_key: The key under which to find the global relations that
            should be used for negative edge sampling.
        neg_sampling_rate: The number of negative (i.e., non-neighbor) edges
            to sample for each real neighborhood edge.
        loss_keys: The keys under which to find the losses that should be
            minimized in this training phase.
        loss_weights: The weights for the losses. If none are specified, losses
            will be weighed equally.
        optimizer: The optmizer to use for loss minimization.
        learning_rate: The learning rate used in the optimization.
        report_interval: How often the loss should be reported during
            training, given in terms of epochs. E.g., with a setting of 5,
            the loss will be reported every 5 epochs.
        kwargs: Additional kwargs that are passed on to the optimizer.

    Attributes:
        loss: The loss constructed from the keys and weights specified above.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        epochs: int = 5,
        batch_size: int = 50,
        batches_per_epoch: int = -1,
        sampling: Literal["standard", "negative_edge"] = "standard",
        edge_rel_key: str = "rel",
        neg_sampling_rate: int = 5,
        loss_keys: list[str] = ["loss"],
        loss_weights: Optional[list[float]] = None,
        optimizer: type = torch.optim.Adam,
        learning_rate: float = 0.01,
        report_interval: int = 5,
        **kwargs,
    ):

        self.name = name
        self.epochs = epochs
        self.batch_size = batch_size
        self.batches_per_epoch = batches_per_epoch
        self.sampling = sampling
        if self.sampling not in ["standard", "negative_edge"]:
            raise ValueError(f"Unknown sampling option {self.sampling}.")
        self.edge_rel_key = edge_rel_key
        self.neg_sampling_rate = neg_sampling_rate

        if not loss_keys:
            raise ValueError("Training phase requires at least one loss key")
        else:
            self.loss_keys = loss_keys

        if loss_weights is None:
            self.loss_weights = list(np.ones(len(self.loss_keys)))
        else:
            self.loss_weights = loss_weights

        self._loss: Optional[pdloss.Loss] = None

        self.optimizer = optimizer
        self.learning_rate = learning_rate
        self.report_interval = report_interval
        self.kwargs = kwargs

        if not issubclass(self.optimizer, torch.optim.Optimizer):
            raise ValueError(
                f"{self.optimizer} is not a valid PyTorch optimizer."
            )

    @property
    def loss(self) -> pdloss.Loss:
        if self._loss is None:
            raise exceptions.LossNotDeterminedError(
                """Attempted to access loss before it was determined."""
            )
        else:
            return self._loss

    @loss.setter
    def loss(self, ls: pdloss.Loss):
        self._loss = ls

    def _determine_loss(self, loss_dict: dict[str, pdloss.Loss]) -> None:
        if len(self.loss_keys) == 1:
            lk = self.loss_keys[0]
            if lk not in loss_dict:
                raise KeyError(f"Invalid loss key {lk}.")
            self.loss = copy.deepcopy(loss_dict[lk])
        else:
            losses: list[pdloss.Loss] = []
            for lk in self.loss_keys:
                if not lk in loss_dict:
                    raise KeyError(f"Invalid loss key {lk}.")
                else:
                    losses.append(copy.deepcopy(loss_dict[lk]))
            self.loss = pdloss.CompoundLoss(
                losses=losses,
                weights=self.loss_weights,
                name=self.name,
            )


RelOrRelDict = Union[relations.Relations, dict[str, relations.Relations]]
LossOrLossDict = Union[pdloss.Loss, dict[str, pdloss.Loss]]

_ParametricDR = TypeVar("_ParametricDR", bound="ParametricDR")


class ParametricDR(utils._ReprMixin):
    """A general parametric dimensionality reduction routine.

    Args:
        model: The PyTorch :class:`torch.nn.module` whose parameters are
            optimized during training.
        in_dim: The numer of dimensions of the input data, used to construct a
            default model in case none is specified. If a dataset is specified
            at instantiation, the correct value for this parameter will be
            inferred from the data dimensions.
        out_dim: The number of output dimensions (i.e., the dimensionality of
            the embedding).
        hidden_dims: Dimensions of hidden layers for the default fully
            connected model that is created if no model is specified.
        derived_data: A dictionary of :class:`paradime.dr.DerivedData`
            instances. These entries are computed before training, either
            before or after the global relations, depending on the options in
            the entries.
        global_relations: A single :class:`paradime.relations.Relations`
            instance or a dictionary with multiple
            :class:`paradime.relations.Relations` instances. Global relations
            are calculated once for the whole dataset before training.
        batch_relations: A single :class:`paradime.relations.Relations`
            instance or a dictionary with multiple
            :class:`paradime.relations.Relations` instances. Batch relations
            are calculated during training for each batch and are compared to
            an appropriate subset of the global relations by a
            :class:`paradime.loss.RelationLoss`.
        losses: A single :class:`paradime.loss.Loss` instance or a dictionary
            with multiple :class:`paradime.loss.Loss` instances. These losses
            are accessed by the training phases via the respective keys.
        training_defaults: A :class:`paradime.dr.TrainingPhase` object with
            settings that override the default values of all other training
            phases. This parameter is useful to avoid having to repeatedly
            set parameters to the same non-default value across training
            phases. Defaults can also be specified after isntantiation using
            the :meth:`set_training_deafults` class method.
        training_phases: A single :class:`paradime.dr.TrainingPhase` object or
            a list of :class:`paradime.dr.TrainingPhase` objects defining the
            training phases to be run. Training phases can also be added
            after instantiation using the :meth:`add_training_phase` class
            method.
        use_cuda: Whether or not to use the GPU for training.
        verbose: Verbosity flag. This setting overrides all verbosity settings
            of relations, transforms and/or losses used within the parametric
            dimensionality reduction.

    Attributes:
        device: The device on which the model is allocated (depends on the
            value specified for ``use_cuda``).
    """

    def __init__(
        self,
        model: Optional[torch.nn.Module] = None,
        in_dim: Optional[int] = None,
        out_dim: int = 2,
        hidden_dims: list[int] = [100, 50],
        derived_data: Optional[dict[str, DerivedData]] = None,
        global_relations: Optional[RelOrRelDict] = None,
        batch_relations: Optional[RelOrRelDict] = None,
        losses: Optional[LossOrLossDict] = None,
        training_defaults: TrainingPhase = TrainingPhase(),
        training_phases: Optional[list[TrainingPhase]] = None,
        use_cuda: bool = False,
        verbose: bool = False,
    ):

        self.verbose = verbose

        self._dataset: Optional[Dataset] = None
        self._dataset_registered = False

        if not derived_data:
            self.derived_data: dict[str, DerivedData] = {}
        else:
            self.derived_data = derived_data

        self.model: torch.nn.Module

        if model is None:
            if isinstance(in_dim, int):
                self.model = models.FullyConnectedEmbeddingModel(
                    in_dim,
                    out_dim,
                    hidden_dims,
                )
            else:
                raise ValueError(
                    "A value for 'in_dim' must be given if no model "
                    "is specified."
                )
        else:
            self.model = model

        if isinstance(global_relations, relations.Relations):
            self.global_relations = {"rel": global_relations}
        elif global_relations is not None:
            self.global_relations = global_relations
        else:
            self.global_relations = {}
        for k in self.global_relations:
            self.global_relations[k]._set_verbosity(self.verbose)

        self.global_relation_data: dict[str, relationdata.RelationData] = {}
        self._global_relations_computed = False

        if isinstance(batch_relations, relations.Relations):
            self.batch_relations = {"rel": batch_relations}
        elif batch_relations is not None:
            self.batch_relations = batch_relations
        else:
            self.batch_relations = {}
        for k in self.batch_relations:
            self.batch_relations[k]._set_verbosity(self.verbose)

        if isinstance(losses, pdloss.Loss):
            self.losses = {"loss": losses}
        elif losses is not None:
            self.losses = losses
        else:
            raise ValueError("No losses specified.")

        self.training_defaults = training_defaults

        self.training_phases: list[TrainingPhase] = []
        if training_phases is not None:
            for tp in training_phases:
                self.add_training_phase(training_phase=tp)

        self.use_cuda = use_cuda
        if use_cuda:
            if torch.cuda.is_available():
                self.model.cuda()
            else:
                raise exceptions.CUDANotAvailableError()

        self.trained = False

    @property
    def dataset(self) -> Dataset:
        if isinstance(self._dataset, Dataset):
            return self._dataset
        else:
            raise exceptions.NoDatasetRegisteredError(
                "Attempted to access a dataset, but none was registered."
            )

    @property
    def device(self) -> torch.device:
        device = torch.device("cpu")
        for p in self.model.parameters():
            device = p.device
            break
        return device

    def __call__(self, X: TensorLike) -> torch.Tensor:

        return self.embed(X)

    @classmethod
    def from_spec(
        cls: Type[_ParametricDR],
        file_or_spec: Union[str, dict],
        model: Optional[models.Model] = None,
    ) -> _ParametricDR:
        """Creates a ParaDime routine from a specification file or dictionary.

        Args:
            file_or_spec: The filename of a YAML or JSON file with a ParaDime
                specification, or a specification in dictionary form.
            model: The PyTorch :class:`torch.nn.module` to be used as the model.
        """

        spec = utils.parsing.validate_spec(file_or_spec)

        derived_data_spec = spec.get("derived data", {})
        derived_entries = _derived_entries_from_spec(derived_data_spec)

        relations_spec = spec.get("relations", {})
        g_rels, b_rels = _relations_from_spec(relations_spec)

        losses_spec = spec.get("losses", {})
        losses = _losses_from_spec(losses_spec)

        tp_spec = spec.get("training phases", {})
        training_phases = _training_phases_from_spec(tp_spec)

        dr = cls(
            model=model,
            derived_data=derived_entries,
            global_relations=g_rels,
            batch_relations=b_rels,
            losses=losses,
            training_phases=training_phases,
        )

        return dr

    def _call_model_method_by_name(
        self,
        method_name: str,
        X: TensorLike,
    ) -> torch.Tensor:

        X = utils.convert.to_torch(X)

        if self.use_cuda:
            X = X.cuda()

        if not hasattr(self.model, method_name):
            raise AttributeError(f"Model has no {method_name} method.")
        elif not callable(getattr(self.model, method_name)):
            raise ValueError(
                f"Attribute {method_name} of model is not a callable."
            )

        if self.trained:
            return getattr(self.model, method_name)(X)
        else:
            raise exceptions.NotTrainedError(
                "DR instance is not trained yet. Call 'train' with "
                "appropriate arguments before calling the model."
            )

    @torch.no_grad()
    def apply(
        self, X: TensorLike, method: Optional[str] = None
    ) -> torch.Tensor:
        """Applies the model to input data.

        Applies the model to an input tensor after first switching off
        PyTorch's automatic gradient tracking. This method also ensures that
        the resulting output tensor is on the CPU. The ``method`` parameter
        allows calling of any of the model's methods in this way, but by
        default, the model's ``__call__`` method will be used (which wraps
        around ``forward``.)

        Args:
            X: A numpy array or PyTorch tensor with the input data.
            method: The name of the model method to be applied.
        """

        if method is None:
            method = "__call__"

        return self._call_model_method_by_name(method, X).cpu()

    def embed(self, X: TensorLike) -> torch.Tensor:
        """Embeds data into the learned embedding space using the model's
        ``embed`` method.

        Args:
            X: A numpy array or PyTorch tensor with the data to be embedded.

        Returns:
            A PyTorch tensor with the embedding coordinates for the data.
        """
        return self._call_model_method_by_name("embed", X)

    def classify(self, X: TensorLike) -> torch.Tensor:
        """Classifies data using the model's ``classify`` method.

        Args:
            X: A numpy array or PyTorch tensor with the data to be classified.

        Returns:
            A PyTorch tensor with the predicted class labels for the data.
        """
        return self._call_model_method_by_name("classify", X)

    def set_training_defaults(
        self,
        training_phase: Optional[TrainingPhase] = None,
        epochs: Optional[int] = None,
        batch_size: Optional[int] = None,
        batches_per_epoch: Optional[int] = None,
        sampling: Optional[Literal["standard", "negative_edge"]] = None,
        edge_rel_key: Optional[str] = None,
        neg_sampling_rate: Optional[int] = None,
        loss_keys: Optional[list[str]] = None,
        loss_weights: Optional[list[float]] = None,
        optimizer: Optional[type] = None,
        learning_rate: Optional[float] = None,
        report_interval: Optional[int] = 5,
        **kwargs,
    ) -> None:
        """Sets a parametric dimensionality reduction routine's default
        training parameters.

        This methods accepts either a :class:`paradime.dr.TrainingPhase`
        instance or individual parameters passed with the same keyword syntax
        used by :class:`paradime.dr.TrainingPhase`. The specified default
        parameters will be used instead of the regular defaults when adding
        training phases.

        Args:
            training_phase: A :class:`paradime.dr.TrainingPhase` instance with
                the new default settings. Instead of this, individual
                parameters can also be passed. For a full list of training
                phase settings, see :class:`paradime.dr.TrainingPhase`.
        """
        if training_phase is not None:
            self.training_defaults = copy.deepcopy(training_phase)
        if epochs is not None:
            self.training_defaults.epochs = epochs
        if batch_size is not None:
            self.training_defaults.batch_size = batch_size
        if batches_per_epoch is not None:
            self.training_defaults.batch_size = batches_per_epoch
        if sampling is not None:
            self.training_defaults.sampling = sampling
        if edge_rel_key is not None:
            self.training_defaults.edge_rel_key = edge_rel_key
        if neg_sampling_rate is not None:
            self.training_defaults.neg_sampling_rate = neg_sampling_rate
        if loss_keys is not None:
            self.training_defaults.loss_keys = loss_keys
        if loss_weights is not None:
            self.training_defaults.loss_weights = loss_weights
        if optimizer is not None:
            self.training_defaults.optimizer = optimizer
        if learning_rate is not None:
            self.training_defaults.learning_rate = learning_rate
        if report_interval is not None:
            self.training_defaults.report_interval = report_interval
        if kwargs:
            self.training_defaults.kwargs = {
                **self.training_defaults.kwargs,
                **kwargs,
            }

    def add_training_phase(
        self,
        training_phase: Optional[TrainingPhase] = None,
        name: Optional[str] = None,
        epochs: Optional[int] = None,
        batch_size: Optional[int] = None,
        batches_per_epoch: Optional[int] = None,
        sampling: Optional[Literal["standard", "negative_edge"]] = None,
        edge_rel_key: Optional[str] = None,
        neg_sampling_rate: Optional[int] = None,
        loss_keys: Optional[list[str]] = None,
        loss_weights: Optional[list[float]] = None,
        optimizer: Optional[type] = None,
        learning_rate: Optional[float] = None,
        report_interval: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Adds a single training phase to a parametric dimensionality
        reduction routine.

        This methods accepts either a :class:`paradime.dr.TrainingPhase`
        instance or individual parameters passed with the same keyword syntax
        used by :class:`paradime.dr.TrainingPhase`.

        Args:
            training_phase: A :class:`paradime.dr.TrainingPhase` instance with
                the new default settings. Instead of this, individual
                parameters can also be passed. For a full list of training
                phase settings, see :class:`paradime.dr.TrainingPhase`.

        Raises:
            :class:`paradime.exceptions.UnsupportedConfigurationError`: This
                error is raised if the type of
                :class:`paradime.relation.Relations` is not compatible with the
                sampling option.
        """
        if training_phase is None:
            training_phase = copy.deepcopy(self.training_defaults)
        assert isinstance(training_phase, TrainingPhase)

        if name is not None:
            training_phase.name = name
        if epochs is not None:
            training_phase.epochs = epochs
        if batch_size is not None:
            training_phase.batch_size = batch_size
        if batches_per_epoch is not None:
            training_phase.batches_per_epoch = batches_per_epoch
        if sampling is not None:
            training_phase.sampling = sampling
        if edge_rel_key is not None:
            training_phase.edge_rel_key = edge_rel_key
        if neg_sampling_rate is not None:
            training_phase.neg_sampling_rate = neg_sampling_rate
        if loss_keys is not None:
            training_phase.loss_keys = loss_keys
        if loss_weights is not None:
            training_phase.loss_weights = loss_weights
        if optimizer is not None:
            training_phase.optimizer = optimizer
        if learning_rate is not None:
            training_phase.learning_rate = learning_rate
        if report_interval is not None:
            training_phase.report_interval = report_interval
        if kwargs:
            training_phase.kwargs = {**training_phase.kwargs, **kwargs}

        training_phase._determine_loss(self.losses)

        if isinstance(
            training_phase.loss, (pdloss.RelationLoss, pdloss.CompoundLoss)
        ):
            training_phase.loss._check_sampling_and_relations(
                training_phase.sampling, self.batch_relations
            )

        self.training_phases.append(training_phase)

    def _register_dataset(self, dataset: Data) -> None:
        """Registers a dataset for a parametric dimensionality reduction
        routine.

        Args:
            dataset: The data, passed either as a single numpy array or PyTorch
                tensor, or a dictionary containing multiple arrays and/or
                tensors.
        """
        if self.verbose:
            utils.logging.log("Initializing training dataset.")
        self._dataset = Dataset(dataset)

        self._dataset_registered = True

    def add_data(
        self, data: Mapping[str, Union[TensorLike, DerivedData]]
    ) -> None:
        """Adds data to a parametric dimensionality reduction routine.

        Tensor-like data will be added to registered dataset. If none is
        registered yet, a new one will be created and registered. Derived
        entries will be added to the routine.

        Args:
            data: A dict containing the data tensors or derived dataset entries
                to be added.
        """
        regular_data: dict[str, TensorLike] = {}
        derived_data: dict[str, DerivedData] = {}

        for k, val in data.items():
            if isinstance(val, (np.ndarray, torch.Tensor)):
                regular_data[k] = val
            elif isinstance(val, DerivedData):
                derived_data[k] = val
            else:
                raise ValueError(
                    f"Value for key {k} is not a numpy array or PyTorch tensor."
                )

        if not self._dataset_registered:
            self._register_dataset(regular_data)
        else:
            for k, val in regular_data.items():
                if self.verbose:
                    if k in self.dataset.data:
                        utils.logging.log(
                            f"Overwriting entry '{k}' in dataset."
                        )
                    else:
                        utils.logging.log(f"Adding entry '{k}' to dataset.")
                self.dataset.data[k] = utils.convert.to_torch(val)

        for k, val in derived_data.items():
            if self.verbose:
                if k in self.derived_data:
                    utils.logging.log(
                        f"Overwriting entry '{k}' in derived data."
                    )
                else:
                    utils.logging.log(f"Adding derived data entry '{k}'.")
            self.derived_data[k] = val

    def compute_derived_data(
        self,
        only: Optional[Literal["rel_based", "other"]] = None,
    ) -> None:
        """Computes the derived data entries in the registered dataset.

        After caling this function, the derived entries will be stored as
        regular entries in the routine's dataset.

        Args:
            only: If "rel_based", only those entries are computed that require
                global relations. If "other", all other entries are computed.
                By default (None), all relations are computed.
        """

        if not self._dataset_registered:
            raise exceptions.NoDatasetRegisteredError(
                "Cannot compute data before registering a dataset."
            )

        compute_rel_based = True
        compute_other = True
        if only == "rel_based":
            compute_other = False
        elif only == "other":
            compute_rel_based = False
        elif only is not None:
            raise ValueError(
                "Expected 'rel', 'other', or None for parameter 'only'."
            )

        for k, entry in self.derived_data.items():
            if (
                entry.requires_relations
                and compute_rel_based
                and not self._global_relations_computed
            ):
                raise exceptions.RelationsNotComputedError(
                    "Cannot compute relation-based derived entry before "
                    "computing global relations."
                )
            elif (entry.requires_relations and compute_rel_based) or (
                not entry.requires_relations and compute_other
            ):
                if self.verbose:
                    utils.logging.log(f"Computing derived data entry '{k}'.")

                options = entry.kwargs

                if "out_dim" not in options and entry.func in [_pca, _spectral]:
                    options["out_dim"] = 2

                selector: dict[
                    str,
                    Union[
                        dict[str, torch.Tensor],
                        dict[str, relationdata.RelationData],
                    ],
                ] = {
                    "data": self.dataset.data,
                    "rels": self.global_relation_data,
                }

                args = [selector[i][j] for i, j in entry.type_key_tuples]
                self.add_data({k: entry.func(*args, **entry.kwargs)})

    def compute_global_relations(self, force: bool = False) -> None:
        """Computes the global relations.

        The computed relation data are stored in the instance's
        ``global_relation_data`` attribute.

        Args:
            force: Whether or not to force a new computation, when relations
                have been previously computed for the same instance.
        """

        if not self._dataset_registered:
            raise exceptions.NoDatasetRegisteredError(
                "Cannot compute global relations before registering dataset."
            )
        assert isinstance(self.dataset, Dataset)

        if force or not self._global_relations_computed:

            for k in self.global_relations:
                if self.verbose:
                    utils.logging.log(f"Computing global relations '{k}'.")
                rel = self.global_relations[k]
                self.global_relation_data[k] = rel.compute_relations(
                    self.dataset.data[rel.data_key]
                )

        self._global_relations_computed = True

    def _prepare_loader(
        self, training_phase: TrainingPhase
    ) -> torch.utils.data.DataLoader:

        if not self._dataset_registered:
            raise exceptions.NoDatasetRegisteredError(
                "Cannot prepare loader before registering dataset."
            )
        assert isinstance(self.dataset, Dataset)

        if not self._global_relations_computed:
            raise exceptions.RelationsNotComputedError(
                "Cannot prepare loader before computing global relations."
            )

        if training_phase.sampling == "negative_edge":
            if training_phase.edge_rel_key not in self.global_relation_data:
                raise KeyError(
                    f"Global relations '{training_phase.edge_rel_key}' "
                    "not specified."
                )
            if training_phase.batches_per_epoch == -1:
                num_edges = max(
                    training_phase.batch_size,
                    int(
                        np.ceil(
                            len(self.dataset)
                            / (training_phase.neg_sampling_rate + 1)
                        )
                    ),
                )
            else:
                num_edges = (
                    training_phase.batch_size * training_phase.batches_per_epoch
                )
            edge_dataset = NegSampledEdgeDataset(
                self.dataset,
                self.global_relation_data[training_phase.edge_rel_key],
                training_phase.neg_sampling_rate,
            )
            sampler = torch.utils.data.WeightedRandomSampler(
                edge_dataset.weights, num_samples=num_edges
            )
            dataloader = torch.utils.data.DataLoader(
                edge_dataset,
                batch_size=training_phase.batch_size,
                collate_fn=_collate_edge_batch,
                sampler=sampler,
            )
        else:
            dataset = self.dataset
            dataloader = torch.utils.data.DataLoader(
                dataset, batch_size=training_phase.batch_size, shuffle=True
            )

        return dataloader

    def _prepare_optimizer(
        self, training_phase: TrainingPhase
    ) -> torch.optim.Optimizer:

        optimizer: torch.optim.Optimizer = training_phase.optimizer(
            self.model.parameters(),
            lr=training_phase.learning_rate,
            **training_phase.kwargs,
        )

        return optimizer

    def _prepare_training(self) -> None:
        """Dummy method to inject code between instantiation and training.

        To be overwritten by subclasses. This allows, e.g., to add default
        training phases outside of a subclass's ``__init__`` but before calling
        the instance's :meth:`train` method.
        """
        pass

    def run_training_phase(self, training_phase: TrainingPhase) -> None:
        """Runs a single training phase.

        Args:
            training_phase: A :class:`paradime.dr.TrainingPhase` instance.
        """
        dataloader = self._prepare_loader(training_phase)
        optimizer = self._prepare_optimizer(training_phase)

        device = self.device

        if self.verbose:
            utils.logging.log(
                f"Beginning training phase '{training_phase.name}'."
            )

        for epoch in range(training_phase.epochs):

            batch: dict[str, torch.Tensor]
            for batch in dataloader:

                optimizer.zero_grad()

                loss = training_phase.loss(
                    self.model,
                    self.global_relation_data,
                    self.batch_relations,
                    batch,
                    device,
                )

                loss.backward()
                optimizer.step()

            training_phase.loss.checkpoint()

            if self.verbose and epoch % training_phase.report_interval == 0:
                # TODO: replace by loss reporting mechanism (GH issue #3)
                utils.logging.log(
                    f"Loss after epoch {epoch}: "
                    f"{training_phase.loss.history[-1]}"
                )

            self.trained = True

    def train(self, data: Optional[Data] = None) -> None:
        """Runs all training phases of a parametric dimensionality reduction
        routine.

        data: The training data, passed either as a single numpy array or
            PyTorch tensor, or as a dictionary containing multiple arrays
            and/or tensors.
        """

        if data is None:
            if not self._dataset_registered:
                raise exceptions.NoDatasetRegisteredError(
                    "Attempted to start training, but no data was pre-added "
                    "or passed to 'train'."
                )
        elif isinstance(data, (np.ndarray, torch.Tensor)):
            self.add_data({"main": data})
        else:
            self.add_data(data)

        self._prepare_training()
        self.compute_derived_data(only="other")
        self.compute_global_relations()
        self.compute_derived_data(only="rel_based")

        self.model.train()

        for tp in self.training_phases:
            self.run_training_phase(tp)

        self.model.eval()


def _pca(x: TensorLike, **kwargs):
    return torch.tensor(
        sklearn.decomposition.PCA(n_components=kwargs["out_dim"]).fit_transform(
            x
        ),
        dtype=torch.float,
    )


def _spectral(reldata: relationdata.RelationData, **kwargs):
    spectral = torch.tensor(
        sklearn.manifold.SpectralEmbedding(
            n_components=kwargs["out_dim"],
            affinity="precomputed",
        ).fit_transform(reldata.to_square_array().data),
        dtype=torch.float,
    )
    spectral = (spectral - spectral.mean(dim=0)) / spectral.std(dim=0)
    return spectral


def _derived_entries_from_spec(
    spec: list[dict],
) -> dict[str, DerivedData]:

    df_dict: dict[str, Callable] = {
        "pca": _pca,
        "spectral": _spectral,
    }

    derived_entries = {}
    for entry in spec:
        derived_entries[entry["name"]] = DerivedData(
            df_dict[entry["data func"]],
            entry["keys"],
            **entry.get("options", {}),
        )
    return derived_entries


def _transforms_from_spec(
    spec: list[dict],
) -> list[transforms.RelationTransform]:

    tf_dict: dict[str, type[transforms.RelationTransform]] = {
        "symmetrize": transforms.Symmetrize,
        "normalize": transforms.Normalize,
        "normalize rows": transforms.NormalizeRows,
        "perplexity": transforms.PerplexityBasedRescale,
        "t-dist": transforms.StudentTTransform,
        "connect": transforms.ConnectivityBasedRescale,
    }

    tfs: list[transforms.RelationTransform] = []

    for tfspec in spec:
        tfs.append(tf_dict[tfspec["type"]](**tfspec.get("options", {})))

    return tfs


def _relations_from_spec(
    spec: list[dict],
) -> tuple[dict[str, relations.Relations], dict[str, relations.Relations]]:

    rel_dict: dict[str, type[relations.Relations]] = {
        "precomp": relations.Precomputed,
        "pdist": relations.PDist,
        "neighbor": relations.NeighborBasedPDist,
        "pdistdiff": relations.DifferentiablePDist,
        "fromto": relations.DistsFromTo,
    }

    global_relations: dict[str, relations.Relations] = {}
    batch_relations: dict[str, relations.Relations] = {}

    for entry in spec:
        data_key = entry.get("field", "main")
        options = entry.get("options", {})
        rel_type = rel_dict[entry["type"]]
        if rel_type is not relations.Precomputed:
            options["data_key"] = data_key
        tfs = _transforms_from_spec(entry["transforms"])
        rel = rel_type(
            transform=tfs,
            **options,
        )
        if entry["level"] == "global":
            global_relations[entry["name"]] = rel
        else:
            batch_relations[entry["name"]] = rel

    return global_relations, batch_relations


def _losses_from_spec(spec: list[dict]) -> dict[str, pdloss.Loss]:

    loss_dict: dict[str, type[pdloss.Loss]] = {
        "relation": pdloss.RelationLoss,
        "classification": pdloss.ClassificationLoss,
        "reconstruction": pdloss.ReconstructionLoss,
        "position": pdloss.PositionLoss,
    }

    loss_func_dict: dict[str, Callable] = {
        "mse": torch.nn.MSELoss(),
        "kl div": pdloss.kullback_leibler_div,
        "cross entropy": torch.nn.CrossEntropyLoss(),
        "umap cross entropy": pdloss.cross_entropy_loss,
    }

    losses: dict[str, pdloss.Loss] = {}

    for entry in spec:
        if loss_dict[entry["type"]] == pdloss.RelationLoss:
            methods = entry["keys"].get("methods", ["embed"])
            losses[entry["name"]] = pdloss.RelationLoss(
                loss_function=loss_func_dict[entry["func"]],
                global_relation_key=entry["keys"]["rels"][0],
                batch_relation_key=entry["keys"]["rels"][1],
                embedding_method=methods[0],
            )
        elif loss_dict[entry["type"]] == pdloss.ClassificationLoss:
            methods = entry["keys"].get("methods", ["classify"])
            losses[entry["name"]] = pdloss.ClassificationLoss(
                loss_function=loss_func_dict[entry["func"]],
                data_key=entry["keys"]["data"][0],
                label_key=entry["keys"]["data"][1],
                classification_method=methods[0],
            )
        elif loss_dict[entry["type"]] == pdloss.ReconstructionLoss:
            methods = entry["keys"].get("methods", ["encode", "decode"])
            losses[entry["name"]] = pdloss.ReconstructionLoss(
                loss_function=loss_func_dict[entry["func"]],
                data_key=entry["keys"]["data"][0],
                encoding_method=methods[0],
                decoding_method=methods[1],
            )
        else:
            methods = entry["keys"].get("methods", ["embed"])
            losses[entry["name"]] = pdloss.PositionLoss(
                loss_function=loss_func_dict[entry["func"]],
                data_key=entry["keys"]["data"][0],
                position_key=entry["keys"]["data"][1],
                embedding_method=methods[0],
            )

    return losses


def _training_phases_from_spec(spec: list[dict]) -> list[TrainingPhase]:

    training_phases: list[TrainingPhase] = []

    for entry in spec:
        if "optimizer" in entry:
            optimizer = {"adam": torch.optim.Adam, "sgd": torch.optim.SGD}[
                entry["optimizer"]["type"]
            ]
            optim_options = entry["optimizer"].get("options", {})
        else:
            optimizer = torch.optim.Adam
            optim_options = {}
        tp = TrainingPhase(
            epochs=entry.get("epochs", 5),
            sampling=(
                "negative_edge"
                if entry["sampling"]["type"] == "edge"
                else "standard"
            ),
            loss_keys=entry["loss"]["components"],
            loss_weights=entry["loss"].get("weights"),
            optimizer=optimizer,
            **entry["sampling"].get("options", {}),
            **optim_options,
        )
        training_phases.append(tp)

    return training_phases



"""A simple script to extract the colors of the rects in an svg-file.
"""

import json
import re
import xml.dom.minidom

import numpy as np

PALETTE_SVG_IN = "palette.svg"
PALETTE_JSON_OUT = "palette.json"

RECT_DTYPE = [("x", float), ("y", float), ("fill", "U7")]
NAME_DTYPE = [("x", float), ("y", float), ("name", "U15")]


def make_palette(in_file: str, out_file: str) -> None:

    svg = xml.dom.minidom.parse(in_file)

    # extract rect elements, coordinates and fills
    rect_vals = []
    for rect in svg.getElementsByTagName("rect"):
        match = re.match(r"fill:([^;]*)", rect.getAttribute("style"))
        rect_vals.append(
            (
                float(rect.getAttribute("x")),
                float(rect.getAttribute("y")),
                match.group(1) if match else "#ffffff",
            )
        )
    rect_data = np.array(rect_vals, dtype=RECT_DTYPE)
    # sort fills by y first, then by x
    rect_data.sort(order=["y", "x"])

    # extract text (i.e., tspan) elements and coordinates
    name_vals = []
    for tspan in svg.getElementsByTagName("tspan"):
        name_vals.append(
            (
                float(tspan.getAttribute("x")),
                float(tspan.getAttribute("y")),
                tspan.childNodes[0].nodeValue,
            )
        )
    name_data = np.array(name_vals, dtype=NAME_DTYPE)
    # sort names by y first, then by x
    name_data.sort(order=["y", "x"])

    palette = dict(zip(name_data["name"], rect_data["fill"]))

    with open(out_file, "w") as f:
        json.dump(palette, f)


if __name__ == "__main__":
    make_palette(PALETTE_SVG_IN, PALETTE_JSON_OUT)



# type: ignore

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import re

sys.path.insert(0, os.path.abspath("../.."))
sys.path.insert(0, os.path.abspath("sphinxext"))

from ghlink import make_linkcode_resolve


# -- Project information -----------------------------------------------------

project = "paradime"
copyright = "2022, Andreas Hinterreiter"
author = "Andreas Hinterreiter"


def get_version():
    for line in open("../../paradime/_version.py").readlines():
        mo = re.match(r"^__version__ = [\"']([^\"']*)[\"']", line)
        if mo:
            return mo.group(1)


release = get_version()
version = release

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.linkcode",
    "sphinx_autodoc_typehints",
    "sphinx-favicon",
    "nbsphinx",
    "nbsphinx_link",
]

nbsphinx_execute = "never"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_logo = "../../assets/ParaDime.svg"
html_theme_options = {
    "logo_only": True,
}

# Favicons (as per https://github.com/tcmetzger/sphinx-favicon)

favicons = [
    {
        "rel": "icon",
        "static-file": "favicons/favicon.svg",
        "type": "image/svg+xml",
    },
    {
        "rel": "icon",
        "sizes": "16x16",
        "static-file": "favicons/favicon16.png",
        "type": "image/png",
    },
    {
        "rel": "icon",
        "sizes": "32x32",
        "static-file": "favicons/favicon32.png",
        "type": "image/png",
    },
    {
        "rel": "apple-touch-icon",
        "sizes": "180x180",
        "static-file": "favicons/favicon180.png",
        "type": "image/png",
    },
]

# Mappings for intersphinx
intersphinx_mapping = {
    "python": (f"https://docs.python.org/{sys.version_info.major}", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "sklearn": ("http://scikit-learn.org/stable/", None),
    "torch": ("https://pytorch.org/docs/stable/", None),
}

# The following is used by sphinx.ext.linkcode to provide links to github
linkcode_resolve = make_linkcode_resolve(
    "paradime",
    "https://github.com/einbandi/paradime/blob/"
    "{revision}/{package}/{path}#L{lineno}",
)



"""Type definitions for ParaDime."""

from typing import Callable, Literal, Tuple, Union

import numpy as np
import numpy.typing as npt
import scipy.sparse
import torch

Rels = Union[
    np.ndarray,
    torch.Tensor,
    scipy.sparse.spmatrix,
    Tuple[np.ndarray, np.ndarray],
]

TensorLike = Union[
    np.ndarray,
    torch.Tensor,
]

TypeKeyTuples = list[tuple[Union[Literal["data"], Literal["rels"]], str]]

BinaryTensorFun = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]

IndexList = Union[list[int], npt.NDArray[np.integer], torch.Tensor]



"""Utility functions for ParaDime.

The :mod:`paradime.utils` subpackage includes various modules that implement
utility functions for logging, plotting, and input conversion.
"""

from paradime.utils import convert
from paradime.utils import logging
from paradime.utils import parsing
from paradime.utils import plotting
from paradime.utils import seed


class _ReprMixin:
    """A mixin implementing a simple __repr__."""

    def __repr__(self) -> str:
        lines = []
        for k, v in self.__dict__.items():
            v_str = repr(v)
            v_str = _addindent(v_str, 2)
            lines.append(f"{k}={v_str},")

        main_str = f"{type(self).__name__}("
        if lines:
            main_str += "\n  " + "\n  ".join(lines) + "\n"

        main_str += ")"
        return main_str


def _addindent(s: str, num_spaces: int) -> str:
    lines = s.split("\n")
    # don't do anything for single-line stuff
    if len(lines) == 1:
        return s
    first = lines.pop(0)
    lines = [(num_spaces * " ") + line for line in lines]
    s = "\n".join(lines)
    s = first + "\n" + s
    return s



"""Relation transforms for ParaDime.

The :mod:`paradime.tranforms` module defines various classes used to transform
relations between data points.
"""

import functools
from typing import Callable, Union, Optional, Any

import numba
import numpy as np
import torch
import scipy.optimize
import scipy.sparse

from paradime import relationdata
from paradime import utils


class RelationTransform(utils._ReprMixin):
    """Base class for relation transforms.

    Custom transforms should subclass this class.
    """

    def __init__(self, **kwargs):
        pass

    def __call__(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        return self.transform(reldata)

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        """Applies the transform to input data.

        Args:
            reldata: The :class:`paradime.relationdata.RelationData` instance
                to be transformed.

        Returns:
            A :class:`paradime.relationdata.RelationData` instance containing
                the transformed relation values.
        """

        raise NotImplementedError()

    def _set_verbosity(self, verbose: bool) -> None:
        if hasattr(self, "verbose"):
            self.verbose = verbose


class Identity(RelationTransform):
    """A placeholder identity transform."""

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata


class ToFlatArray(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.FlatRelationArray`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_flat_array()


class ToFlatTensor(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.FlatRelationTensor`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_flat_tensor()


class ToSquareArray(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.SquareRelationArray`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_square_array()


class ToSquareTensor(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.SquareRelationTensor`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_square_tensor()


class ToTriangularArray(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.TriangularRelationArray`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_triangular_array()


class ToTriangularTensor(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.TriangularRelationTensor`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_triangular_tensor()


class ToNeighborTuple(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.NeighborRelationTuple`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_neighbor_tuple()


class ToSparseArray(RelationTransform):
    """Converts the relations to a
    :class:`paradime.relationdata.SparseRelationArray`.
    """

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:
        return reldata.to_sparse_array()


class AdaptiveNeighborhoodRescale(RelationTransform):
    """Rescales relation values for each data point based on its neighbors.

    This is a base class for transformations such as those used by t-SNE
    or UMAP. For each data point, a parameter is fitted by comparing
    kernel-transformed relations to a target value. Once the parameter
    value is found, the kernel function is used to transform the relations.

    Args:
        kernel: The kernel function used to transform the relations. This
            is a callable taking the relation values for a data point,
            along with a parameter.
        find_param: The function used to find the parameter value. This is a
            callable taking the relation values and a fixed value to compare
            the transformed relations against.
        verbose: Verbosity toggle.
    """

    def __init__(
        self,
        kernel: Callable[[np.ndarray, float], Union[float, np.ndarray]],
        find_param: Callable[..., float],
        verbose: bool = False,
        **kwargs,
    ):
        super().__init__()

        self.kernel = kernel
        self.find_param = find_param
        self.verbose = verbose
        self.kwargs = kwargs

        self._param_values: Optional[np.ndarray] = None

    @property
    def param_values(self) -> np.ndarray:
        """The parameter values determined for each data point. Available only
        after calling the transform.
        """
        if self._param_values is not None:
            return self._param_values
        else:
            raise AttributeError(
                "Parameter values only available after calling transform."
            )

    def _set_root_scalar_defaults(self) -> None:
        raise NotImplementedError()

    def _get_comparison_constant(self) -> float:
        raise NotImplementedError()

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        reldata = reldata.to_neighbor_tuple()
        reldata._remove_self_relations()

        self._set_root_scalar_defaults()

        neighbors, relations = reldata.data
        num_pts = len(neighbors)
        self._param_values = np.empty(num_pts, dtype=float)

        if self.verbose:
            utils.logging.log("Calculating probabilities.")

        for i, rels in enumerate(relations):
            beta = self.find_param(
                rels, self._get_comparison_constant(), **self.kwargs
            )
            self._param_values[i] = beta
            relations[i] = self.kernel(rels, beta)

        reldata.data = (neighbors, relations)

        return reldata


class PerplexityBasedRescale(AdaptiveNeighborhoodRescale):
    """Applies a perplexity-based transformation to the relation values.

    The relation values are rescaled using Guassian kernels. For each data
    point, the kernel width is determined by comparing the entropy of the
    relation values to the binary logarithm of the specified perplexity.
    This is the relation transform used by t-SNE.

    Args:
        perplexity: The desired perplexity, which can be understood as
            a smooth measure of nearest neighbors.
        verbose: Verbosity toggle.
        **kwargs: Passed on to :func:`scipy.optimize.root_scalar`, which
            determines the kernel widths. By default, this is set to use a
            bracket of [0.01, 1.] for the root search.
    """

    def __init__(
        self,
        perplexity: float = 30,
        verbose: bool = False,
        **kwargs,  # passed on to root_scalar
    ):
        super().__init__(_p_i, _find_beta, verbose=verbose, **kwargs)

        self.perplexity = perplexity

    @property
    def betas(self) -> np.ndarray:
        return self.param_values

    def _get_comparison_constant(self) -> float:
        return self.perplexity

    def _set_root_scalar_defaults(self) -> None:
        if not self.kwargs:  # check if emtpy
            self.kwargs["bracket"] = [0.01, 1.0]


@numba.jit
def _entropy(dists: np.ndarray, beta: float) -> float:
    x = -(dists**2) * beta
    y = np.exp(x)
    ysum = y.sum()

    if ysum < 1e-50:
        result = -1.0
    else:
        factor = -1 / (np.log(2.0) * ysum)
        result = factor * ((y * x) - (y * np.log(ysum))).sum()

    return result


def _p_i(dists: np.ndarray, beta: float) -> np.ndarray:
    x = -(dists**2) * beta
    y = np.exp(x)
    ysum = y.sum()

    return y / ysum


def _find_beta(dists: np.ndarray, perp: float, **kwargs) -> float:
    return scipy.optimize.root_scalar(
        lambda b: _entropy(dists, b) - np.log2(perp), **kwargs
    ).root


class ConnectivityBasedRescale(AdaptiveNeighborhoodRescale):
    """Applies a connectivity-based transformation to the relation values.

    The relation values are rescaled using shifted Guassian kernels. The
    shift is equal to the closes neighboring data point, and the kernel
    width is set by by comparing the summed kernel values to the binary
    logarithm of the specified number of neighbors. This is the relation
    transform used by UMAP.

    Args:
        n_neighbors: The number of nearest neighbors used to determine the
            kernel widths.
        verbose: Verbosity toggle.
        **kwargs: Passed on to :func:`scipy.optimize.root_scalar`, which
            determines the kernel widths. By default, this is set to use a
            bracket of [10^(-6), 10^6] for the root search.
    """

    def __init__(
        self,
        n_neighbors: float = 15,
        verbose: bool = False,
        **kwargs,  # passed on to root_scalar
    ):
        super().__init__(_exp_k, _find_sigma, verbose=verbose, **kwargs)

        self.n_neighbors = n_neighbors

    @property
    def sigmas(self) -> np.ndarray:
        return self.param_values

    def _get_comparison_constant(self) -> float:
        return self.n_neighbors

    def _set_root_scalar_defaults(self) -> None:
        if not self.kwargs:  # check if emtpy
            self.kwargs["bracket"] = [1.0e-6, 1.0e6]


def _exp_k(dists: np.ndarray, sigma: float) -> np.ndarray:
    x = dists - dists.min()
    return np.exp(-x / sigma)


def _find_sigma(dists: np.ndarray, k: int, **kwargs) -> float:
    return scipy.optimize.root_scalar(
        lambda s: _exp_k(dists, s).sum() - np.log2(float(k)), **kwargs
    ).root


class Symmetrize(RelationTransform):
    """Symmetrizes the relation values.

    Args:
        subtract_product: Specifies which symmetrization routine to use.
            If set to False (default), a matrix M is symmetrized by
            calculating 1/2 * (M + M^T); if set to True, M is symmetrized
            by calculating M + M^T - M * M^T, where '*' is the element-wise
            (Hadamard) product.
    """

    def __init__(self, subtract_product: bool = False):
        super().__init__()
        self.subtract_product = subtract_product

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        if isinstance(
            reldata,
            (
                relationdata.TriangularRelationArray,
                relationdata.TriangularRelationTensor,
            ),
        ):
            return reldata
        else:
            if self.subtract_product:
                symmetrizer = _sym_subtract_product
            else:
                symmetrizer = _sym_plus_only

            if isinstance(reldata, relationdata.NeighborRelationTuple):
                return symmetrizer(reldata.to_sparse_array())
            else:
                return symmetrizer(reldata)


def _sym_plus_only(
    reldata: relationdata.RelationData,
) -> relationdata.RelationData:
    if isinstance(reldata, relationdata.SquareRelationArray):
        reldata.data = 0.5 * (reldata.data + reldata.data.T)
    elif isinstance(reldata, relationdata.SparseRelationArray):
        reldata.data = 0.5 * (reldata.data + reldata.data.transpose())
    elif isinstance(reldata, relationdata.SquareRelationTensor):
        reldata.data = 0.5 * (reldata.data + torch.t(reldata.data))
    else:
        raise TypeError(
            "Expected non-flat :class:`paradime.relationdata.RelationData`."
        )
    return reldata


def _sym_subtract_product(
    reldata: relationdata.RelationData,
) -> relationdata.RelationData:
    if isinstance(reldata, relationdata.SquareRelationArray):
        reldata.data = (
            reldata.data + reldata.data.T - reldata.data * reldata.data.T
        )
    elif isinstance(reldata, relationdata.SparseRelationArray):
        reldata.data = (
            reldata.data
            + reldata.data.transpose()
            - reldata.data.multiply(reldata.data.transpose())
        )
    elif isinstance(reldata, relationdata.SquareRelationTensor):
        reldata.data = (
            reldata.data
            + torch.t(reldata.data)
            - reldata.data * torch.t(reldata.data)
        )
    else:
        raise TypeError(
            "Expected non-flat :class:`paradime.relationdata.RelationData`."
        )
    return reldata


class NormalizeRows(RelationTransform):
    """Normalizes the relation values for each data point separately."""

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        if isinstance(reldata, relationdata.TriangularRelationArray):
            reldata = reldata.to_square_array()
            reldata.data /= reldata.data.sum(axis=1, keepdims=True)
        elif isinstance(reldata, relationdata.TriangularRelationTensor):
            reldata = reldata.to_square_tensor()
            reldata.data /= reldata.data.sum(dim=1, keepdim=True)
        elif isinstance(reldata, relationdata.SquareRelationArray):
            reldata.data /= reldata.data.sum(axis=1, keepdims=True)
        elif isinstance(reldata, relationdata.SquareRelationTensor):
            reldata.data /= reldata.data.sum(dim=1, keepdim=True)
        elif isinstance(reldata, relationdata.SparseRelationArray):
            norm_factors = 1 / np.repeat(
                np.array(reldata.data.sum(axis=1)), reldata.data.getnnz(axis=1)
            )
            reldata.data = scipy.sparse.csr_matrix(
                (norm_factors, reldata.data.nonzero())
            ).multiply(reldata.data)
        elif isinstance(reldata, relationdata.NeighborRelationTuple):
            neighbors, relations = reldata.data
            reldata.data = (
                neighbors,
                relations / relations.sum(axis=1, keepdims=True),
            )
        else:
            raise TypeError(
                "Expected non-flat :class:`paradime.relationdata.RelationData`."
            )
        return reldata


class Normalize(RelationTransform):
    """Normalizes all relations at once."""

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        if isinstance(
            reldata,
            (
                relationdata.TriangularRelationArray,
                relationdata.TriangularRelationTensor,
            ),
        ):
            reldata.data /= 2 * reldata.data.sum()
        elif isinstance(
            reldata,
            (
                relationdata.SquareRelationArray,
                relationdata.SquareRelationTensor,
                relationdata.SparseRelationArray,
                relationdata.FlatRelationTensor,
                relationdata.FlatRelationArray,
            ),
        ):
            reldata.data /= reldata.data.sum()
        elif isinstance(reldata, relationdata.NeighborRelationTuple):
            neighbors, relations = reldata.data
            reldata.data = (neighbors, relations / relations.sum())
        else:
            raise TypeError(
                "Expected :class:`paradime.relationdata.RelationData`."
            )
        return reldata


class ZeroDiagonal(RelationTransform):
    """Sets all self-relations to zero."""

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        if isinstance(
            reldata,
            (
                relationdata.TriangularRelationArray,
                relationdata.TriangularRelationTensor,
            ),
        ):
            return reldata
        elif isinstance(reldata, relationdata.SquareRelationArray):
            np.fill_diagonal(reldata.data, 0.0)
        elif isinstance(reldata, relationdata.SquareRelationTensor):
            reldata.data.fill_diagonal_(0.0)
        elif isinstance(reldata, relationdata.SparseRelationArray):
            reldata.data.setdiag(0.0)
        elif isinstance(reldata, relationdata.NeighborRelationTuple):
            reldata._remove_self_relations()
        else:
            raise TypeError(
                "Expected non-flat :class:`paradime.relationdata.RelationData`."
            )
        return reldata


class StudentTTransform(RelationTransform):
    """Transforms relations based on Student's t-distribution.

    Args:
        alpha: Degrees of freedom of the distribution. This can either be
            a float or a PyTorch tensor. Alpha can be optimized together
            with the DR model in a :class:`paradime.dr.ParametricDR` by
            setting it to one of the model's additional parameters.
    """

    def __init__(self, alpha: Union[float, torch.Tensor]):
        self.alpha = alpha

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        if isinstance(
            reldata,
            (
                relationdata.SquareRelationTensor,
                relationdata.TriangularRelationTensor,
                relationdata.FlatRelationTensor,
            ),
        ):
            reldata.data = reldata.data.pow(2.0)
            reldata.data = reldata.data / self.alpha
            reldata.data = reldata.data + 1.0
            # TODO: fix RuntimeError due to in-place operation below
            reldata.data = reldata.data.pow(-(self.alpha + 1.0) / 2.0)
        elif isinstance(
            reldata,
            (
                relationdata.SquareRelationArray,
                relationdata.TriangularRelationArray,
                relationdata.FlatRelationArray,
            ),
        ):
            reldata.data = np.power(
                1.0 + reldata.data**2.0 / self.alpha,
                -(self.alpha + 1.0) / 2.0,
            )
        elif isinstance(reldata, relationdata.SparseRelationArray):
            return self.transform(reldata.to_square_array())
        elif isinstance(reldata, relationdata.NeighborRelationTuple):
            neighbors, rels = reldata.data
            rels = np.power(
                1.0 + rels**2.0 / self.alpha, -(self.alpha + 1.0) / 2.0
            )
            reldata.data = (neighbors, rels)
        else:
            raise TypeError(
                "Expected :class:`paradime.relationdata.RelationData`."
            )
        return reldata


class ModifiedCauchyTransform(RelationTransform):
    """Transforms relations based on a modified Cauchy distribution.

    This transform applies a modified Cauchy distribution function
    to the relations. The distribution's parameters ``a`` and ``b`` are
    determined from the parameters ``min_dist`` and ``spread`` by fitting a
    smooth approximation of an offset exponential decay.

    Args:
        min_dist: Effective minimum distance of points if the transformed
            relations were to be used for calculating an embedding.
        spread: Effective scale of the points if the tranformed relations
            were to be used for calculating an embedding.
        a: Parameter to define the distribution directly. It can be optimized
            together with the DR model in a :class:`ParametricDR` by setting
            it to one of the model's additional parameters.
        b: Parameter to define the distribution directly. It can be optimized
            together with the DR model in a :class:`ParametricDR` by setting
            it to one of the model's additional parameters.
    """

    def __init__(
        self,
        min_dist: float = 0.1,
        spread: float = 1.0,
        a: Optional[Union[float, torch.Tensor]] = None,
        b: Optional[Union[float, torch.Tensor]] = None,
    ):
        self.min_dist = min_dist
        self.spread = spread

        self.a: Union[float, torch.Tensor]
        self.b: Union[float, torch.Tensor]

        if a is None or b is None:
            self.a, self.b = _find_ab_params(self.spread, self.min_dist)
        if a is not None:
            self.a = a
        if b is not None:
            self.b = b

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        if isinstance(
            reldata,
            (
                relationdata.SquareRelationTensor,
                relationdata.TriangularRelationTensor,
                relationdata.FlatRelationTensor,
            ),
        ):
            reldata.data = reldata.data.pow(2.0 * self.b)
            reldata.data = torch.pow(1.0 + self.a * reldata.data, -1.0)
        elif isinstance(
            reldata,
            (
                relationdata.SquareRelationArray,
                relationdata.TriangularRelationArray,
                relationdata.FlatRelationArray,
            ),
        ):
            reldata.data = np.power(1.0 + reldata.data ** (2.0 * self.b), -1.0)
        elif isinstance(reldata, relationdata.SparseRelationArray):
            return self.transform(reldata.to_square_array())
        elif isinstance(reldata, relationdata.NeighborRelationTuple):
            neighbors, rels = reldata.data
            rels = np.power(1.0 + rels ** (2.0 * self.b), -1.0)
            reldata.data = (neighbors, rels)
        else:
            raise TypeError(
                "Expected :class:`paradime.relationdata.RelationData`."
            )
        return reldata


@functools.cache
def _find_ab_params(spread: float, min_dist: float) -> tuple[float, float]:
    def curve(x, a, b):
        return 1.0 / (1.0 + a * x ** (2 * b))

    xv = np.linspace(0, spread * 3, 300)
    yv = np.zeros(xv.shape)
    yv[xv < min_dist] = 1.0
    yv[xv >= min_dist] = np.exp(-(xv[xv >= min_dist] - min_dist) / spread)
    params, _ = scipy.optimize.curve_fit(curve, xv, yv)
    return params[0], params[1]


class Functional(RelationTransform):
    """Applies a function to the relation data.

    By default, this transform applies a given function to the :attr:`data`
    attribute of the :class:`paradime.relationdata.RelationData` instance in
    place and returns the transformed instance. This assumes that the transform
    does not change the data in a way that is incompatible with the
    :class:`paradime.relationdata.RelationData` subclass. The transform can
    also be applied to the whole :class:`paradime.relationdata.RelationData`
    instance by setting ``in_place`` to False. In this case, the output is that
    of the given function.

    Args:
        f: Function to be applied to the relations.
        in_place: Toggles whether the function is applied to the :attr:`data`
            attribute of the :class:`paradime.relationdata.RelationData` object
            (default), or to the :class:`paradime.relationdata.RelationData`
            itself.
        check_valid: Toggles whether a check for the transformed relation
            data's validity is performed. If ``in_place`` is set to False, no
            checks are performed regardless of this parameter.
    """

    def __init__(
        self,
        f: Callable[..., Any],
        in_place: bool = True,
        check_valid: bool = False,
    ):
        self.f = f
        self.in_place = in_place
        self.check_valid = check_valid

    def transform(
        self,
        reldata: relationdata.RelationData,
    ) -> relationdata.RelationData:

        if self.in_place:
            reldata.data = self.f(reldata.data)
            if self.check_valid:
                try:
                    type(reldata)(reldata.data)  # type: ignore
                except:
                    raise ValueError(
                        "Transformed relation data no longer compatible "
                        f"with type {type(reldata).__name__}."
                    )
            return reldata
        else:
            return self.f(reldata)
